"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import traceback
from functools import wraps
from time import sleep

from botocore.exceptions import ClientError

from syndicate.exceptions import ResourceProcessingError
from syndicate.commons.log_helper import get_logger

_LOG = get_logger(__name__)

LOG_NOT_FOUND_ERROR = 'log_not_found_error'

DEFAULT_RETRY_TIMEOUT_SEC = 35
DEFAULT_RETRY_TIMEOUT_STEP = 3


def apply_methods_decorator(decorator):
    # todo after applying this decorator static methods do not work if they
    #  are invoked from an instance of a class instead of a class.
    # self.some_static_method(1, 2, 3)  # won't work
    def decorate(cls):
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)):
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls

    return decorate


def retry(retry_timeout=DEFAULT_RETRY_TIMEOUT_SEC,
          retry_timeout_step=DEFAULT_RETRY_TIMEOUT_STEP):
    """ Decorator for retry on specified exceptions.

    :type handler_func: func
    :param handler_func: function which will be decorated
    :param retry_timeout: retry timeout in seconds
    :param retry_timeout_step: step of the retry
    """

    def decorator(handler_func):
        @wraps(handler_func)
        def wrapper(*args, **kwargs):
            """ Wrapper func."""
            retry_exceptions = [
                'ThrottlingException',
                'LimitExceededException',
                'ProvisionedThroughputExceededException',
                'TooManyRequestsException',
                'ConflictException',
                'An error occurred (InvalidParameterValueException) when '
                'calling the CreateEventSourceMapping operation',
                'An error occurred (InvalidParameterValueException) when '
                'calling the UpdateEventSourceMapping operation',
                'An error occurred (ResourceInUseException) when '
                'calling the UpdateEventSourceMapping operation'
                'An error occurred (InvalidParameterValueException) when '
                'calling the CreateCluster operation',
                'An error occurred (InvalidParameterValue) when '
                'calling the CreateQueue operation',
                'An error occurred (SubnetGroupInUseFault) when calling '
                'the DeleteSubnetGroup operation',
                'The role defined for the function cannot be assumed by Lambda',
                'An error occurred (ResourceConflictException) when calling'
                ' the AddPermission operation: The statement id',
                'NoSuchUpload',
                'Throttling',
                'Please add Lambda as a Trusted Entity',
                'UpdateFunctionConfiguration',
                'PutScalingPolicy',
                'RegisterScalableTarget',
                'TopicArn can not be None',
                'DeleteRole',
                'Max attempts exceeded',
                'UpdateGatewayResponse',
                'Cannot delete, found existing JobQueue relationship',
                'Cannot delete, resource is being modified',
                'Please try again',
                'An error occurred (ConcurrentModificationException) when '
                'calling the CreateDataSource operation: Schema is currently '
                'being altered',
                'An error occurred (ConcurrentModificationException) when '
                'calling the CreateResolver operation: Schema is currently '
                'being altered',
                'An error occurred (ConcurrentModificationException) when '
                'calling the UpdateResolver operation: Schema is currently '
                'being altered',
                'An error occurred (InvalidArgument) when calling the '
                'PutBucketNotificationConfiguration operation',
                'Too Many Requests'
            ]
            resource_not_found_error_codes = [
                'NoSuchEntity',
                'ResourceNotFoundException',
                'StateMachineDoesNotExist',
                'ClusterNotFoundFault',
                'InvalidInstanceID.NotFound',
                'IncorrectInstanceState'

            ]
            last_ex = None
            for each in range(1, retry_timeout, retry_timeout_step):
                try:
                    return handler_func(*args, **kwargs)
                except ClientError as e:
                    retry_flag = False
                    for exc in retry_exceptions:
                        if exc in str(e):
                            _LOG.warning(f'Retry on {handler_func.__name__}. '
                                         f'Error: {str(e)}')
                            _LOG.debug(
                                f'Parameters: {str(args)}, {str(kwargs)}')
                            # set to debug, we need it only in the logs file
                            _LOG.debug(
                                f'Traceback:\n {traceback.format_exc()}')
                            retry_flag = True
                            break
                    if not retry_flag:
                        error_code = e.response['Error']['Code']
                        if (kwargs.get(LOG_NOT_FOUND_ERROR) and error_code in
                           resource_not_found_error_codes):
                            _LOG.error(f'Error occurred: {e}')
                            _LOG.error(
                                f'Traceback:\n {traceback.format_exc()}')
                        else:
                            _LOG.debug(f'Error occurred: {e}')
                            _LOG.debug(
                                f'Traceback:\n {traceback.format_exc()}')
                        raise e
                    last_ex = e
                    sleep(each)

            if last_ex:
                raise ResourceProcessingError(
                    f"Maximum retries reached for function "
                    f"'{handler_func.__name__}' due to "
                    f"'{type(last_ex).__name__}': "
                    f"'{str(last_ex)}'") from last_ex
        return wrapper
    return decorator



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from troposphere import Ref, iam

from syndicate.connection.iam_connection import IAMConnection
from syndicate.commons.log_helper import get_user_logger
from .cf_resource_converter import CfResourceConverter
from ..cf_transform_utils import (to_logic_name, iam_role_logic_name, is_arn,
                                  iam_managed_policy_logic_name,
                                  iam_instance_profile_logic_name)

_LOG = get_user_logger()


class CfIamRoleConverter(CfResourceConverter):

    def convert(self, name, meta):
        role = iam.Role(iam_role_logic_name(name))
        role.RoleName = name
        role.Policies = []
        role.ManagedPolicyArns = self._prepare_policy_arns(meta)
        role.AssumeRolePolicyDocument = \
            self._build_assume_role_policy_document(meta)
        self.template.add_resource(role)

        instance_profile = meta.get('instance_profile')
        if instance_profile and str(instance_profile).lower() == 'true':
            self._convert_instance_profile(role_name=name)

    def _prepare_policy_arns(self, meta):
        custom_policies = meta.get('custom_policies') or []
        predefined_policies = meta.get('predefined_policies') or []
        policy_arns = []
        for policy in predefined_policies:
            iam_service = self.resources_provider.iam()
            policy_arn = iam_service.iam_conn.get_policy_arn(
                policy, policy_scope='AWS')
            if not policy_arn:
                _LOG.warning(f'AWS-managed policy with name \'{policy}\' '
                             f'not found. Skipping...')
                continue
            policy_arns.append(policy_arn)
        for policy in custom_policies:
            policy_arns.append(Ref(iam_managed_policy_logic_name(policy)))
        return policy_arns

    @staticmethod
    def _build_assume_role_policy_document(meta):
        allowed_accounts = meta.get('allowed_accounts', [])
        principal_service = meta.get('principal_service')
        external_id = meta.get('external_id')
        trusted_relationships = meta.get('trusted_relationships')
        if not trusted_relationships:
            trusted_relationships = IAMConnection.empty_trusted_relationships()
        if allowed_accounts:
            trusted_accounts = IAMConnection.set_allowed_account(
                allowed_accounts, external_id, 'create')
            trusted_relationships['Statement'].append(trusted_accounts)
        if principal_service:
            trusted_services = IAMConnection.set_allowed_service(
                principal_service, 'create')
            trusted_relationships['Statement'].append(trusted_services)
        return trusted_relationships

    def _convert_instance_profile(self, role_name):
        logic_name = iam_instance_profile_logic_name(role_name)
        instance_profile = iam.InstanceProfile(logic_name)
        instance_profile.InstanceProfileName = role_name
        instance_profile.Roles = [Ref(iam_role_logic_name(role_name))]
        self.template.add_resource(instance_profile)

    @staticmethod
    def convert_inline_policy(role, policy_name, policy_document):
        logic_name = to_logic_name('IAMPolicy', policy_name)
        policy = iam.PolicyType(logic_name)
        policy.PolicyDocument = policy_document
        policy.PolicyName = policy_name
        policy.Roles = [role if isinstance(role, str) else role.ref(), ]
        return policy

    @staticmethod
    def attach_managed_policy(role, policy):
        if is_arn(policy):
            if policy not in role.ManagedPolicyArns:
                role.ManagedPolicyArns.append(policy)
        else:
            ref = Ref(policy)
            if ref not in role.ManagedPolicyArns:
                role.ManagedPolicyArns.append(ref)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import os
from pathlib import Path

from syndicate.exceptions import InvalidValueError
from syndicate.commons.log_helper import get_logger
from syndicate.core import ProjectState
from syndicate.core.generators import (_touch, _mkdir,
                                       _write_content_to_file)
from syndicate.core.generators.contents import (_get_lambda_default_policy,
                                                JAVA_ROOT_POM_TEMPLATE,
                                                SRC_MAIN_JAVA, FILE_POM,
                                                CHANGELOG_TEMPLATE,
                                                GITIGNORE_CONTENT,
                                                README_TEMPLATE)
from syndicate.core.groups import (RUNTIME_JAVA, RUNTIME_NODEJS,
                                   RUNTIME_PYTHON, RUNTIME_DOTNET)

_LOG = get_logger(__name__)

SLASH_SYMBOL = '/'
FOLDER_LAMBDAS = 'lambdas'
FOLDER_COMMONS = 'commons'
FILE_README = 'README.md'
FILE_DEPLOYMENT_RESOURCES = 'deployment_resources.json'
FILE_CHANGELOG = 'CHANGELOG.md'
FILE_GITIGNORE = '.gitignore'


def generate_project_structure(project_name, project_path):
    try:
        if not os.path.exists(project_path):
            raise InvalidValueError(
                f"Path '{project_path}' you have provided does not exist"
            )

        full_project_path = os.path.join(project_path, project_name) \
            if (project_path[-1] != SLASH_SYMBOL) \
            else project_path + project_name

        _mkdir(path=full_project_path,
               fault_message='Folder {} already exists. \nOverride the '
                             'project? [y/n]: '.format(full_project_path))

        path_to_readme = os.path.join(full_project_path, FILE_README)
        _touch(path_to_readme)
        readme_template = README_TEMPLATE.replace('project_name', project_name)
        _write_content_to_file(file=path_to_readme,
                               content=readme_template)

        default_lambda_policy = _get_lambda_default_policy()
        _write_content_to_file(file=os.path.join(full_project_path,
                                                 FILE_DEPLOYMENT_RESOURCES),
                               content=default_lambda_policy)
        ProjectState.generate(project_name=project_name,
                              project_path=full_project_path)

        _write_content_to_file(os.path.join(full_project_path, FILE_CHANGELOG),
                               CHANGELOG_TEMPLATE)
        _write_content_to_file(os.path.join(full_project_path, FILE_GITIGNORE),
                               GITIGNORE_CONTENT)
        _LOG.info('Project {} folder has been successfully created.'.format(
            project_name))
    except Exception as e:
        _LOG.exception(str(e))


def _generate_python_project_hierarchy(full_project_path, project_name=None):
    _mkdir(os.path.join(full_project_path, FOLDER_LAMBDAS), exist_ok=True)


def _generate_java_project_hierarchy(project_name, full_project_path):
    pom_path = os.path.join(full_project_path, FILE_POM)
    if not os.path.exists(pom_path):
        _LOG.info(f"Generating pom inside {full_project_path}")
        _touch(path=pom_path)
        pom_content = JAVA_ROOT_POM_TEMPLATE.replace('{project_name}',
                                                     project_name)
        _write_content_to_file(file=pom_path, content=pom_content)
        _LOG.info(f"Pom file was generated by path: {pom_path}")
    else:
        _LOG.info(f"Pom file inside {full_project_path} already exists")
    _mkdir(Path(full_project_path, SRC_MAIN_JAVA), exist_ok=True)


def _generate_nodejs_project_hierarchy(full_project_path, project_name=None):
    _mkdir(os.path.join(full_project_path, FOLDER_LAMBDAS), exist_ok=True)
    _mkdir(os.path.join(full_project_path, FOLDER_COMMONS), exist_ok=True)


def _generate_dotnet_project_hierarchy(full_project_path, project_name=None):
    _mkdir(os.path.join(full_project_path, FOLDER_LAMBDAS), exist_ok=True)


PROJECT_PROCESSORS = {
    RUNTIME_JAVA: _generate_java_project_hierarchy,
    RUNTIME_NODEJS: _generate_nodejs_project_hierarchy,
    RUNTIME_PYTHON: _generate_python_project_hierarchy,
    RUNTIME_DOTNET: _generate_dotnet_project_hierarchy
}



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.core.generators.deployment_resources.dynamodb_generator \
    import DynamoDBGenerator, DynamoDBGlobalIndexGenerator, \
    DynamoDBAutoscalingGenerator
from syndicate.core.generators.deployment_resources.s3_bucket_generator \
    import S3Generator
from syndicate.core.generators.deployment_resources.api_gateway_generator \
    import ApiGatewayGenerator, ApiGatewayResourceGenerator, \
    ApiGatewayResourceMethodGenerator
from syndicate.core.generators.deployment_resources.iam_policy_generator \
    import IAMPolicyGenerator
from syndicate.core.generators.deployment_resources.iam_role_generator \
    import IAMRoleGenerator
from syndicate.core.generators.deployment_resources.kinesis_stream_generator \
    import KinesisStreamGenerator
from syndicate.core.generators.deployment_resources.sns_topic_generator \
    import SNSTopicGenerator
from syndicate.core.generators.deployment_resources. \
    step_function_activity_generator import StepFunctionActivityGenerator
from syndicate.core.generators.deployment_resources.base_generator import \
    BaseConfigurationGenerator
from syndicate.core.generators.deployment_resources.ec2_instance_generator \
    import EC2InstanceGenerator
from syndicate.core.generators.deployment_resources.sqs_queue_generator \
    import SQSQueueGenerator
from syndicate.core.generators.deployment_resources.sns_application_generator \
    import SNSApplicationGenerator
from syndicate.core.generators.deployment_resources. \
    cognito_user_pool_generator import CognitoUserPoolGenerator
from syndicate.core.generators.deployment_resources. \
    batch_compenv_generator import BatchCompenvGenerator
from syndicate.core.generators.deployment_resources.batch_jobqueue_generator \
    import BatchJobqueueGenerator
from syndicate.core.generators.deployment_resources.batch_jobdef_generator \
    import BatchJobdefGenerator
from syndicate.core.generators.deployment_resources. \
    cloudwatch_alarm_generator import CloudWatchAlarmGenerator
from syndicate.core.generators.deployment_resources.step_function_generator \
    import StepFunctionGenerator
from syndicate.core.generators.deployment_resources.cognito_federated_generator \
    import CognitoFederatedPoolGenerator
from syndicate.core.generators.deployment_resources. \
    cloudwatch_event_rule_generator import CloudwatchEventRuleGenerator
from syndicate.core.generators.deployment_resources.docdb_cluster_generator \
    import DocumentDBClusterGenerator
from syndicate.core.generators.deployment_resources.docdb_instance_generator \
    import DocumentDBInstanceGenerator
from syndicate.core.generators.deployment_resources.dax_cluster_generator \
    import DaxClusterGenerator
from syndicate.core.generators.deployment_resources.eventbridge_rule_generator \
    import EventBridgeRuleGenerator
from syndicate.core.generators.deployment_resources.\
    web_socket_api_gateway_generator import WebSocketApiGatewayGenerator



import json
from pathlib import Path, PurePath

from syndicate.exceptions import ResourceMetadataError, AbortedError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import DYNAMO_TABLE_TYPE, IAM_ROLE, \
    COGNITO_USER_POOL_TYPE, APPSYNC_CONFIG_FILE_NAME, \
    APPSYNC_JS_RESOLVER_CODE_DEFAULT_FILE_NAME, \
    APPSYNC_VTL_RESOLVER_REQ_MT_DEFAULT_FILE_NAME, \
    APPSYNC_VTL_RESOLVER_RESP_MT_DEFAULT_FILE_NAME, APPSYNC_SRC_FOLDER
from syndicate.core.generators import _write_content_to_file, \
    _read_content_from_file, _mkdir, _touch
from syndicate.core.generators.contents import \
    _generate_syncapp_js_resolver_code, _generate_syncapp_vtl_resolver_req_mt, \
    _generate_syncapp_vtl_resolver_resp_mt
from syndicate.core.generators.deployment_resources import \
    BaseConfigurationGenerator
from click import confirm as click_confirm


APPSYNC_RESOLVERS_DIR = 'resolvers'
APPSYNC_FUNCTIONS_DIR = 'functions'
DEFAULT_FUNC_VTL_RMT_VERSION = '2018-05-29'

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


class AppSyncConfigurationGenerator(BaseConfigurationGenerator):
    """Contains common method for AppSync meta generators"""

    def __init__(self, **kwargs):
        from syndicate.core import CONFIG
        path_to_project = CONFIG.project_path
        self.appsync_name = kwargs.pop('api_name')
        self.appsync_path = PurePath(path_to_project, APPSYNC_SRC_FOLDER,
                                     self.appsync_name).as_posix()
        self.appsync_config = self._get_appsync_config()
        self.region = CONFIG.region
        super().__init__(**kwargs)

    def _get_appsync_config(self):
        path_to_config = PurePath(self.appsync_path,
                                  APPSYNC_CONFIG_FILE_NAME).as_posix()
        if not Path(path_to_config).is_file():
            raise ResourceMetadataError(
                f"Config file of the AppSync '{self.appsync_name}' not found. "
                f"Please make sure that the AppSync exists.")
        return json.loads(_read_content_from_file(path_to_config))

    def _save_config(self):
        _write_content_to_file(
            PurePath(self.appsync_path, APPSYNC_CONFIG_FILE_NAME).as_posix(),
            json.dumps(self.appsync_config, indent=2))


class AppSyncDataSourceGenerator(AppSyncConfigurationGenerator):
    CONFIGURATION = {
        'name': str,
        'description': None,
        'type': None,
        'resource_name': None,
        'region': None,
        'service_role_name': None,
        'lambda_config': None,
        'dynamodb_config': None
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def write(self):
        data_sources = self.appsync_config.get('data_sources', [])
        current_ds_name = self._dict.get('name')
        for data_source in data_sources:
            if current_ds_name == data_source['name']:
                message = (f"Data source with the name '{current_ds_name}' "
                           f"already exists.")
                if click_confirm(f"{message} Overwrite?"):
                    _LOG.warning(
                        f"Overwriting data source '{current_ds_name}'")
                    data_sources.remove(data_source)
                else:
                    _LOG.warning(
                        f"Skipping data source '{current_ds_name}'")
                    raise AbortedError
        data_sources.append(self._resolve_configuration())
        self.appsync_config['data_sources'] = data_sources
        self._save_config()

    def _resolve_configuration(self, defaults_dict=None):
        region = self._dict.pop('region', None) or self.region
        data_source_type = self._dict.get('type')
        if data_source_type != 'NONE':
            self._validate_resource_existence(
                self._dict.get('service_role_name'),
                IAM_ROLE)
        match data_source_type:
            case 'AWS_LAMBDA':
                lambda_name = self._dict.pop('resource_name')
                self._validate_lambda_existence(lambda_name)
                self._dict['lambda_config'] = {
                    'lambda_name': lambda_name,
                    'aws_region': region
                }
            case 'AMAZON_DYNAMODB':
                table_name = self._dict.pop('resource_name')
                self._validate_resource_existence(table_name,
                                                  DYNAMO_TABLE_TYPE)
                self._dict['dynamodb_config'] = {
                    'table_name': table_name,
                    'aws_region': region
                }

        return super()._resolve_configuration()


class AppSyncFunctionGenerator(AppSyncConfigurationGenerator):
    CONFIGURATION = {
        'name': str,
        'description': str,
        'data_source_name': str,
        'runtime': str,
        'code_path': None,
        'function_version': None,
        'request_mapping_template_path': None,
        'response_mapping_template_path': None,
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def write(self):
        data_sources = self.appsync_config.get('data_sources', [])
        current_ds_name = self._dict.get('data_source_name')
        error_message = (
            f"Data source '{current_ds_name}' not found in the SyncApp "
            f"'{self.appsync_name}' definition.")
        for data_source in data_sources:
            if current_ds_name == data_source['name']:
                self._dict['data_source_type'] = data_source['type']
                error_message = None
                break
        if error_message:
            raise ResourceMetadataError(error_message)

        current_func_name = self._dict['name']
        functions = self.appsync_config.get('functions', [])
        for function in functions:
            if function['name'].lower() == current_func_name.lower():
                message = (f"The function with the name '{current_func_name}' "
                           f"already exists.")
                if click_confirm(f"{message} Overwrite?"):
                    _LOG.warning(
                        f"Overwriting function '{current_func_name}'")
                    functions.remove(function)
                else:
                    _LOG.warning(f"Skipping function'{current_func_name}'")
                    raise AbortedError

        functions.append(self._resolve_configuration())
        self.appsync_config['functions'] = functions
        self._save_config()

    def _resolve_configuration(self, defaults_dict=None):
        runtime = self._dict.get('runtime')
        internal_path_to_func = PurePath(
            APPSYNC_FUNCTIONS_DIR,
            self._dict.get('name').lower()).as_posix()

        path_to_func = PurePath(
            self.appsync_path,
            internal_path_to_func).as_posix()

        paths_to_code = PurePath(
            path_to_func,
            APPSYNC_JS_RESOLVER_CODE_DEFAULT_FILE_NAME).as_posix()

        paths_to_req_mapping_template = PurePath(
            path_to_func,
            APPSYNC_VTL_RESOLVER_REQ_MT_DEFAULT_FILE_NAME).as_posix()
        paths_to_resp_mapping_template = PurePath(
            path_to_func,
            APPSYNC_VTL_RESOLVER_RESP_MT_DEFAULT_FILE_NAME).as_posix()

        if not Path(path_to_func).exists():
            _mkdir(path_to_func)
        if runtime == 'JS':
            Path(paths_to_req_mapping_template).unlink(missing_ok=True)
            Path(paths_to_resp_mapping_template).unlink(missing_ok=True)
            self._dict['code_path'] = PurePath(
                internal_path_to_func,
                APPSYNC_JS_RESOLVER_CODE_DEFAULT_FILE_NAME).as_posix()
            code_content = _generate_syncapp_js_resolver_code()
            _touch(paths_to_code)
            _write_content_to_file(paths_to_code, code_content)
        if runtime == 'VTL':
            Path(paths_to_code).unlink(missing_ok=True)
            req_mapping_template = _generate_syncapp_vtl_resolver_req_mt(
                self._dict['data_source_type'])
            resp_mapping_template = _generate_syncapp_vtl_resolver_resp_mt(
                self._dict['data_source_type'])

            self._dict['function_version'] = DEFAULT_FUNC_VTL_RMT_VERSION
            self._dict['request_mapping_template_path'] = PurePath(
                internal_path_to_func,
                APPSYNC_VTL_RESOLVER_REQ_MT_DEFAULT_FILE_NAME).as_posix()
            self._dict['response_mapping_template_path'] = PurePath(
                internal_path_to_func,
                APPSYNC_VTL_RESOLVER_RESP_MT_DEFAULT_FILE_NAME).as_posix()

            _touch(paths_to_req_mapping_template)
            _write_content_to_file(paths_to_req_mapping_template,
                                   req_mapping_template)

            _touch(paths_to_resp_mapping_template)
            _write_content_to_file(paths_to_resp_mapping_template,
                                   resp_mapping_template)

        return super()._resolve_configuration()


class AppSyncResolverGenerator(AppSyncConfigurationGenerator):
    CONFIGURATION = {
        'kind': str,
        'type_name': str,
        'field_name': str,
        'data_source_name': None,
        'runtime': str,
        'pipeline_config': None,
        'code_path': None,
        'request_mapping_template_path': None,
        'response_mapping_template_path': None,
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def write(self):
        if self._dict['kind'] == 'UNIT':
            data_sources = self.appsync_config.get('data_sources', [])
            current_ds_name = self._dict.get('data_source_name')
            error_message = (
                f"Data source '{current_ds_name}' not found in the SyncApp "
                f"'{self.appsync_name}' definition.")
            for data_source in data_sources:
                if current_ds_name == data_source['name']:
                    self._dict['data_source_type'] = data_source['type']
                    error_message = None
                    break
            if error_message:
                raise ResourceMetadataError(error_message)
        elif self._dict['kind'] == 'PIPELINE':
            if (functions := self.appsync_config.get('functions', [])) is None:
                functions = []
            if (curr_functions := self._dict.get('function_name', [])) is None:
                curr_functions = []
            absent_funcs = []

            for func_name in set(curr_functions):
                func_exist = False
                for func_conf in functions:
                    if func_conf['name'].lower() == func_name.lower():
                        func_exist = True
                        break
                if not func_exist:
                    absent_funcs.append(func_name)
            if absent_funcs:
                raise ResourceMetadataError(
                    f"The next function/s '{absent_funcs}' not found in the "
                    f"SyncApp '{self.appsync_name}' definition.")
            self._dict['pipeline_config'] = {
                'functions': curr_functions
            }
        resolvers = self.appsync_config.get('resolvers', [])
        current_type_name = self._dict.get('type_name')
        current_field_name = self._dict.get('field_name')
        for resolver in resolvers:
            if (resolver['type_name'].lower() == current_type_name.lower()
                    and
               resolver['field_name'].lower() == current_field_name.lower()):
                message = (f"The resolver for the type '{current_type_name}' "
                           f"and field '{current_field_name}' already exists.")
                if click_confirm(f"{message} Overwrite?"):
                    _LOG.warning(
                        f"Overwriting resolver for the type "
                        f"'{current_type_name}' and field "
                        f"'{current_field_name}'")
                    resolvers.remove(resolver)
                else:
                    _LOG.warning(
                        f"Skipping resolver for the type "
                        f"'{current_type_name}' and field "
                        f"'{current_field_name}'")
                    raise AbortedError

        resolvers.append(self._resolve_configuration())
        self.appsync_config['resolvers'] = resolvers
        self._save_config()

    def _resolve_configuration(self, defaults_dict=None):
        runtime = self._dict.get('runtime')
        internal_path_to_field = PurePath(
            APPSYNC_RESOLVERS_DIR, self._dict.get('type_name').lower(),
            self._dict.get('field_name').lower()).as_posix()

        path_to_field = PurePath(
            self.appsync_path,
            internal_path_to_field).as_posix()

        paths_to_code = PurePath(
            path_to_field,
            APPSYNC_JS_RESOLVER_CODE_DEFAULT_FILE_NAME).as_posix()

        paths_to_req_mapping_template = PurePath(
            path_to_field,
            APPSYNC_VTL_RESOLVER_REQ_MT_DEFAULT_FILE_NAME).as_posix()
        paths_to_resp_mapping_template = PurePath(
            path_to_field,
            APPSYNC_VTL_RESOLVER_RESP_MT_DEFAULT_FILE_NAME).as_posix()

        if not Path(path_to_field).exists():
            _mkdir(path_to_field)
        if runtime == 'JS':
            Path(paths_to_req_mapping_template).unlink(missing_ok=True)
            Path(paths_to_resp_mapping_template).unlink(missing_ok=True)
            self._dict['code_path'] = PurePath(
                internal_path_to_field,
                APPSYNC_JS_RESOLVER_CODE_DEFAULT_FILE_NAME).as_posix()
            code_content = _generate_syncapp_js_resolver_code()
            _touch(paths_to_code)
            _write_content_to_file(paths_to_code, code_content)
        if runtime == 'VTL':
            Path(paths_to_code).unlink(missing_ok=True)
            data_source_type = (
                self._dict['data_source_type']
                if self._dict['kind'] == 'UNIT' else 'PIPELINE')
            req_mapping_template = _generate_syncapp_vtl_resolver_req_mt(
                data_source_type)
            resp_mapping_template = _generate_syncapp_vtl_resolver_resp_mt(
                data_source_type)

            self._dict['request_mapping_template_path'] = PurePath(
                internal_path_to_field,
                APPSYNC_VTL_RESOLVER_REQ_MT_DEFAULT_FILE_NAME).as_posix()
            self._dict['response_mapping_template_path'] = PurePath(
                internal_path_to_field,
                APPSYNC_VTL_RESOLVER_RESP_MT_DEFAULT_FILE_NAME).as_posix()

            _touch(paths_to_req_mapping_template)
            _write_content_to_file(paths_to_req_mapping_template,
                                   req_mapping_template)

            _touch(paths_to_resp_mapping_template)
            _write_content_to_file(paths_to_resp_mapping_template,
                                   resp_mapping_template)

        return super()._resolve_configuration()


class AppSyncAuthorizationGenerator(AppSyncConfigurationGenerator):
    CONFIGURATION = {
        'primary_auth_type': None,
        'extra_auth_types': None,
        'lambda_authorizer_config': None,
        'user_pool_config': None,
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def write(self):
        primary_auth = self.appsync_config.get('primary_auth_type')
        self._dict['current_primary_auth'] = primary_auth
        if self._dict['type'] == 'primary':
            if primary_auth:
                message = "Primary authorization already exists."
                if click_confirm(f"{message} Overwrite?"):
                    _LOG.warning(f"Overwriting primary authorization")
                    self.appsync_config.pop('lambda_authorizer_config', None)
                    self.appsync_config.pop('user_pool_config', None)
                else:
                    _LOG.warning(f"Skipping primary authorization creation")
                    raise AbortedError

        extra_auth = self.appsync_config.get('extra_auth_types', [])
        self._dict['current_extra_auth'] = extra_auth
        if self._dict['type'] == 'extra':
            if not primary_auth:
                raise ResourceMetadataError(
                    'Primary authorization is mandatory, please configure it '
                    'first'
                )
            new_extra_auth_type = self._dict['auth_type']
            for auth_type in extra_auth:
                if auth_type['authentication_type'] == new_extra_auth_type:
                    message = (f"Extra authorization '{new_extra_auth_type}' "
                               f"already exists.")
                    if click_confirm(f"{message} Overwrite?"):
                        _LOG.warning(f"Overwriting extra authorization")
                    else:
                        _LOG.warning(
                            f"Skipping extra authorization creation")
                        raise AbortedError

        self.appsync_config.update(self._resolve_configuration())
        self._save_config()

    def _resolve_configuration(self, defaults_dict=None):
        authorizer_type = self._dict.pop('type')
        authentication_type = self._dict.pop('auth_type')
        authentication_config = self._resolve_authentication_configuration(
            authentication_type
        )
        if authorizer_type == 'primary':
            primary_auth_type = authentication_config['authentication_type']
            lambda_config = \
                authentication_config.get('lambda_authorizer_config')
            cognito_config = authentication_config.get('user_pool_config')

            for extra_auth in self._dict['current_extra_auth']:
                if primary_auth_type == extra_auth['authentication_type']:
                    raise ResourceMetadataError(
                        f"'{primary_auth_type}' can't be configured as "
                        f" the primary authorization because it is already "
                        f"configured as an extra authorization provider")

            self._dict['primary_auth_type'] = primary_auth_type
            if lambda_config:
                self._dict['lambda_authorizer_config'] = lambda_config
            if cognito_config:
                cognito_config['default_action'] = 'DENY'
                self._dict['user_pool_config'] = cognito_config
        elif authorizer_type == 'extra':
            new_auth_type = authentication_config['authentication_type']

            if new_auth_type == self._dict['current_primary_auth']:
                raise ResourceMetadataError(
                    f"'{new_auth_type}' can't be configured as "
                    f"an extra authorization provider because it is already "
                    f"configured as the primary authorization")

            current_extra_auth = self._dict.pop('current_extra_auth', [])
            for extra_auth in current_extra_auth:
                if extra_auth['authentication_type'] == new_auth_type:
                    current_extra_auth.remove(extra_auth)

            current_extra_auth.append(authentication_config)
            self._dict['extra_auth_types'] = current_extra_auth

        return super()._resolve_configuration()

    def _resolve_authentication_configuration(self, auth_type: str) -> dict:
        region = self._dict.pop('region', None) or self.region

        match auth_type:
            case 'API_KEY':
                return {'authentication_type': 'API_KEY'}
            case 'AWS_IAM':
                return {'authentication_type': 'AWS_IAM'}
            case 'AWS_LAMBDA':
                lambda_name = self._dict.pop('resource_name')
                self._validate_lambda_existence(lambda_name)

                return {
                    'authentication_type': 'AWS_LAMBDA',
                    'lambda_authorizer_config': {
                        'authorizer_result_ttl_in_seconds': 300,
                        'resource_name': lambda_name,
                        'aws_region': region
                    }
                }
            case 'AMAZON_COGNITO_USER_POOLS':
                cup_name = self._dict.pop('resource_name')
                self._validate_resource_existence(cup_name,
                                                  COGNITO_USER_POOL_TYPE)

                return {
                    'authentication_type': 'AMAZON_COGNITO_USER_POOLS',
                    'user_pool_config': {
                        'resource_name': cup_name,
                        'aws_region': region
                    }
                }



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import re

from syndicate.exceptions import ParameterError, ResourceProcessingError, InvalidValueError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.conf.processor import GLOBAL_AWS_SERVICES
from typing import TypeVar, Optional, Iterable
from datetime import datetime

T = TypeVar('T')

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


def validate_params(name, meta, required_params):
    existing_parameters = list(meta.keys())
    parameters_string = ', '.join(required_params)
    existing_parameters_string = ', '.join(existing_parameters)
    for each in required_params:
        if each not in existing_parameters:
            raise ParameterError(
                f"All required parameters must be specified! "
                f"Resource: '{name}' "
                f"Required parameters: '{parameters_string}'. "
                f"Given parameters: '{existing_parameters_string}'"
            )


def validate_known_params(name: str, act_params: Iterable,
                          known_params: Iterable):
    unknown_params = []
    for param in act_params:
        if param not in known_params:
            unknown_params.append(param)

    if unknown_params:
        raise ParameterError(
            f"Unknown parameter/s detected in the '{name}' configuration. "
            f"Unknown parameter/s: {unknown_params}. "
            f"Must be one of: {known_params}"
        )


def validate_date(name, date_str):
    """
    Checks if the provided date in `date_str` is in ISO 8601 or Unix timestamp format
    """
    try:
        datetime.fromisoformat(date_str)
        unix_timestamp = int(date_str)
        datetime.utcfromtimestamp(unix_timestamp)
        return date_str
    except Exception as e:
        raise InvalidValueError(
            f"Invalid date format: '{date_str}'. Resource: '{name}' "
            f"Error message: {e}"
        )


def check_region_available(region_name, available_regions, res_meta=None):
    if region_name in available_regions:
        return True
    if res_meta:
        res_type = res_meta['resource_type']
        raise ResourceProcessingError(
            f"Region '{region_name}' isn't available for resource '{res_type}'."
        )
    else:
        raise ResourceProcessingError(
            f"Region '{region_name}' isn't available."
        )


def create_args_for_multi_region(args, available_regions):
    from syndicate.core import CONFIG
    new_region_args = []
    for arg_set in args:
        name = arg_set['name']
        meta = arg_set['meta']
        region = meta.get('region')
        if region is None:
            item = arg_set.copy()
            item['region'] = CONFIG.region
            new_region_args.append(item)
        elif isinstance(region, str):
            if region == 'all':
                for each in available_regions:
                    item = arg_set.copy()
                    item['region'] = each
                    new_region_args.append(item)
            else:
                if check_region_available(region, available_regions, meta):
                    item = arg_set.copy()
                    item['region'] = region
                    new_region_args.append(item)
        elif isinstance(region, list):
            for each in region:
                if check_region_available(each, available_regions, meta):
                    item = arg_set.copy()
                    item['region'] = each
                    new_region_args.append(item)
        else:
            raise InvalidValueError(
                f"Invalid value region: '{region}'. Resource: '{name}'."
            )
    return new_region_args


def chunks(l, n):
    for i in range(0, len(l), n):
        yield l[i:i + n]


def resolve_dynamic_identifier(to_replace, resource_meta):
    """
    Replaces keys from 'to_replace' with values from it inside json built
    from 'resource_meta'
    :type to_replace: dict
    :type resource_meta: dict
    """
    raw_json = json.dumps(resource_meta)
    for name, value in to_replace.items():
        raw_json = raw_json.replace(name, value)
    return json.loads(raw_json)


def detect_unresolved_aliases(resource_meta: dict) -> None:
    """
    Extract unresolved aliases in the given meta string.
    :param resource_meta: Input dict.
    """
    pattern = r'\$\{([^}]+)\}'
    if missing_aliases := list(set(re.findall(pattern,
                                              json.dumps(resource_meta)))):
        placeholders = [f'${{{alias}}}' for alias in missing_aliases]
        USER_LOG.warning(
            'Unresolved alias placeholders are present in the meta, which may '
            'lead to errors in resource processing. Please check the '
            'syndicate aliases and resource definitions. '
            f'Unresolved placeholders: {placeholders}')


def build_description_obj(response, name, meta):
    resource_type = meta['resource_type']
    obj = {
        'resource_name': name,
        'resource_meta': meta,
        'description': response
    }
    if resource_type not in GLOBAL_AWS_SERVICES:
        from syndicate.core import CONFIG
        obj['resource_meta']['region'] = meta.get('region', CONFIG.region)
    return obj


def assert_required_params(required_params_names, all_params):
    """
    Raises error if there is at least one missing parameter.
    :param required_params_names:
    :param all_params:
    :return:
    """
    missing = [param for param in required_params_names if
               param not in all_params.keys()]
    if missing:
        raise ParameterError(f"Missing required parameters: '{missing}'")


def assert_possible_values(iterable: list, possible: list):
    if not set(iterable).issubset(set(possible)):
        message = f'Incorrect values in given iterable: {iterable}. ' \
                  f'Must be a subset of these: {possible}'
        _LOG.error(message)
        raise InvalidValueError(message)


def filter_dict_by_shape(d, shape):
    new_d = {}
    for attribute, value in shape.items():
        if value is None:
            new_d[attribute] = d.get(attribute)

        if isinstance(value, list):
            new_d[attribute] = filter_list_by_shape(d.get(attribute), value)

        if isinstance(value, dict):
            new_d[attribute] = filter_dict_by_shape(d.get(attribute), value)
    return new_d


def filter_list_by_shape(lst, shape):
    if not shape[0] or not lst:
        return lst

    new_lst = []
    if isinstance(shape[0], dict):
        for d in lst:
            new_lst.append(filter_dict_by_shape(d, shape[0]))

    return new_lst


def if_updated(new: T, old: T) -> Optional[T]:
    """
    If `new` differs from `old` it `new` will be returned.
    If it does not, None is returned. They must be comparable
    """
    return new if new != old else None



import re
from abc import ABC, abstractmethod
from typing import Union
from collections.abc import Iterable

from syndicate.exceptions import InvalidTypeError, InternalError, \
    NotImplementedError

NAMED_S3_URI_PATTERN = r'^(?P<proto>s3:\/\/)?(?:(?P<name>[0-9a-z\-]+)' \
                       r'(?:\/)?)(?P<key>(?<=\/)(?:[0-9a-z\-_]+(?:\/)?)+)?$'
S3_PATTERN_GROUP_NAMES = ('proto', 'name', 'key')


class AbstractViewDigest(ABC):
    """
    Abstract parsing view class.
    """
    @abstractmethod
    def parse(self, value: str):
        if not isinstance(value, str):
            raise InvalidTypeError(
                'Value to be parsed must be of string type.'
            )


class RegexViewDigest(AbstractViewDigest):
    def __init__(self):
        self._expression = None
        self._groups: Iterable = tuple()

    def parse(self, value: str) -> dict:
        """
        Regex digestion of the incoming string value.
        :return: dict | keys = self.groups
        """
        super(self.__class__, self).parse(value)
        if not self.expression:
            raise InternalError('No expression has been assigned.')
        match = self.expression.match(value)
        return match.groupdict() if match else dict()

    @property
    def expression(self):
        return self._expression

    @expression.setter
    def expression(self, pattern: str):
        """
        Prepares a regex expression, based on the provided pattern.
        The pattern is assigned, only if the expression contains
        requires groups, if there are any.
        """
        try:
            pending = re.compile(pattern)
        except re.error as exception:
            raise exception

        if any(each for each in self.groups if each not in pending.groupindex):
            raise InternalError(
                f"Pattern '{pattern}', must include named groups "
                f"'{', '.join(self.groups)}'"
            )
        self._expression = pending

    @property
    def groups(self) -> Iterable:
        return self._groups

    @groups.setter
    def groups(self, other: Iterable):
        """
        Sets up required expression groups
        as an iterable collection of strings.
        """
        if not isinstance(other, Iterable) and any(
            each for each in other if not isinstance(each,str)
        ):
            raise InvalidTypeError(
                'Required groups must be an iterable and contain only string '
                'elements.'
            )
        self._groups = other


class AbstractBucketView(ABC):
    """
    Abstract bucket view.
    """
    class BucketViewRuntimeError(RuntimeError):
        """
        The runtime parent bucket view error class.
        """

    def __init__(self):
        self._raw = None
        self._digest = None

    @property
    @abstractmethod
    def name(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def raw(self):
        return self._raw

    @raw.setter
    @abstractmethod
    def raw(self, value: str) -> None:
        """
        Abstract raw value setter, which provides
        argument type verification.
        :raises: InvalidTypeError
        """
        if not self.digest:
            raise self.BucketViewRuntimeError('View digest hasn\'t been set.')
        if not isinstance(value, str):
            raise InvalidTypeError('Raw value must be of string type.')

    @property
    def digest(self):
        return self._digest

    @digest.setter
    def digest(self, other: AbstractViewDigest):
        if isinstance(other, AbstractViewDigest):
            self._digest = other
        else:
            raise InvalidTypeError(
                'View digest-parser must be of AbstractViewDigest type.'
            )


class URIBucketView(AbstractBucketView):

    class InvalidS3URIException(AbstractBucketView.BucketViewRuntimeError):
        """
        An error class, meant to thrown once an invalid S3 URI is inputted.
        """

    def __init__(self):
        """
        Initializes a bucket view, with an empty parsed-value maintainer.
        """
        super(self.__class__, self).__init__()
        self._parsed = None

    @property
    def raw(self) -> Union[str, None]:
        return self._raw

    @raw.setter
    def raw(self, value: str):
        """
        Installs a raw url value, which alters the
        parsed-value maintainer in runtime.
        """
        super(self.__class__, self.__class__).raw.fset(self, value)
        self._raw = value
        self._parsed = self.digest.parse(value)

    @property
    def name(self) -> Union[str, None]:
        """
        Returns the name of a bucket, retrieving netloc of the url.
        """
        value = self._parsed.get('name', '') if self._parsed else None
        return '' if value is None else value

    @property
    def key(self) -> str:
        """
        Returns the key-object path compound.
        """
        value = self._parsed.get('key', '') if self._parsed else None
        return '' if value is None else value




"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from boto3 import client

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger(__name__)


@apply_methods_decorator(retry())
class DaxConnection:
    """DynamoDB DAX connection clas"""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('dax', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new DAX connection.')

    def create_cluster(self, cluster_name: str, node_type: str,
                       replication_factor: str, iam_role_arn: str,
                       subnet_group_name: str = None,
                       cluster_endpoint_encryption_type: str = 'TLS',
                       security_group_ids: list = None,
                       parameter_group_name: str = None,
                       availability_zones: list = None,
                       tags: list[dict] = None):
        params = dict(
            ClusterName=cluster_name,
            NodeType=node_type,
            ReplicationFactor=replication_factor,
            IamRoleArn=iam_role_arn,
            SubnetGroupName=subnet_group_name,
            SecurityGroupIds=security_group_ids,
            ParameterGroupName=parameter_group_name,
            AvailabilityZones=availability_zones,
            ClusterEndpointEncryptionType=cluster_endpoint_encryption_type,
            Tags=tags
        )
        params = {key: value for key, value in params.items() if value}
        try:
            return self.client.create_cluster(**params)
        except self.client.exceptions.ClusterAlreadyExistsFault as e:
            _LOG.warning(f'Cluster \'{cluster_name}\' already exists')
            return

    def describe_cluster(self, cluster_name):
        try:
            response = self.client.describe_clusters(
                ClusterNames=[cluster_name],
            )
            return response['Clusters'][0]
        except self.client.exceptions.ClusterNotFoundFault:
            _LOG.warning(f'Cluster \'{cluster_name}\' not found')
            return

    def delete_cluster(self, cluster_name, log_not_found_error=True):
        """
        log_not_found_error parameter is needed for proper log handling in the
        retry decorator
        """
        return self.client.delete_cluster(ClusterName=cluster_name)

    def create_subnet_group(self, subnet_group_name, subnet_ids):
        params = dict(
            SubnetGroupName=subnet_group_name,
            SubnetIds=subnet_ids
        )
        return self.client.create_subnet_group(**params)

    def describe_subnet_group(self, subnet_group_name: str):
        try:
            return self.client.describe_subnet_groups(
                SubnetGroupNames=[subnet_group_name]
            )['SubnetGroups'][0]
        except self.client.exceptions.SubnetGroupNotFoundFault:
            _LOG.warning(f'Subnet group \'{subnet_group_name}\' not found')
            return

    @retry(retry_timeout=120, retry_timeout_step=10)
    def delete_subnet_group(self, subnet_group_name: str):
        try:
            self.client.delete_subnet_group(
                SubnetGroupName=subnet_group_name
            )
        except self.client.exceptions.SubnetGroupNotFoundFault:
            _LOG.warning(f'Subnet group with name '
                         f'\'{subnet_group_name}\' not found')
            return



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json

from boto3 import client
from botocore.exceptions import ClientError

from syndicate.exceptions import InvalidValueError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


@apply_methods_decorator(retry())
class SqsConnection(object):
    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.region = region
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.client = client('sqs', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new SQS connection.')

    def create_queue(self, queue_name, delay_seconds=None,
                     maximum_message_size=None, message_retention_period=None,
                     policy=None, receive_message_wait_time_seconds=None,
                     redrive_policy=None, visibility_timeout=None,
                     kms_master_key_id=None,
                     kms_data_key_reuse_period_seconds=None, fifo_queue=False,
                     content_based_deduplication=None, tags=None):
        attributes = dict()
        if fifo_queue:
            attributes['FifoQueue'] = str(fifo_queue)
        params = dict(QueueName=queue_name)
        if delay_seconds is not None:
            if delay_seconds < 0 or delay_seconds > 900:
                raise InvalidValueError(
                    'Delay seconds for queue must be between 0 and 900 seconds'
                )
            attributes['DelaySeconds'] = str(delay_seconds)
        if maximum_message_size is not None:
            if maximum_message_size < 1024 or maximum_message_size > 262144:
                raise InvalidValueError(
                    'Maximum message size must be between 1024 and 262144 bytes'
                )
            attributes['MaximumMessageSize'] = str(maximum_message_size)
        if message_retention_period is not None:
            if message_retention_period < 60 or message_retention_period > 1209600:
                raise InvalidValueError(
                    'Message retention size must be between 60 and 1209600 seconds'
                )
            attributes['MessageRetentionPeriod'] = str(
                message_retention_period)
        if policy:
            if isinstance(policy, dict):
                policy = json.dumps(policy)
            attributes['Policy'] = policy
        if receive_message_wait_time_seconds is not None:
            if receive_message_wait_time_seconds < 0 or receive_message_wait_time_seconds > 20:
                raise InvalidValueError(
                    'Receive message wait time must be between 0 and 20 seconds'
                )
            attributes[
                'ReceiveMessageWaitTimeSeconds'] = str(
                receive_message_wait_time_seconds)
        if redrive_policy:
            if isinstance(redrive_policy, dict):
                redrive_policy = json.dumps(redrive_policy)
            attributes['RedrivePolicy'] = redrive_policy
        if visibility_timeout is not None:
            if visibility_timeout < 0 or visibility_timeout > 43200:
                raise InvalidValueError(
                    'Visibility timeout must be between 0 and 43200 seconds'
                )
            attributes['VisibilityTimeout'] = str(visibility_timeout)
        if kms_master_key_id:
            attributes['KmsMasterKeyId'] = kms_master_key_id
        if kms_data_key_reuse_period_seconds is not None:
            if kms_data_key_reuse_period_seconds < 60 or kms_data_key_reuse_period_seconds > 86400:
                raise InvalidValueError(
                    'KMS key reuse period must be between 60 and 86400 seconds'
                )
            attributes[
                'KmsDataKeyReusePeriodSeconds'] = str(
                kms_data_key_reuse_period_seconds)
        if content_based_deduplication:
            attributes[
                'ContentBasedDeduplication'] = str(content_based_deduplication)
        params['Attributes'] = attributes
        if tags:
            params['tags'] = tags
        return self.client.create_queue(**params)

    def delete_queue(self, queue_url, log_not_found_error=True):
        """
        log_not_found_error parameter is needed for proper log handling in the
        retry decorator
        """
        self.client.delete_queue(QueueUrl=queue_url)

    def list_queues(self, url_prefix):
        response = self.client.list_queues(QueueNamePrefix=url_prefix)
        return response.get('QueueUrls', list())

    def get_queue_url(self, queue_name, account_id):
        try:
            response = self.client.get_queue_url(QueueName=queue_name,
                                                 QueueOwnerAWSAccountId=account_id)
            return response['QueueUrl']
        except ClientError as e:
            if 'AWS.SimpleQueueService.NonExistentQueue' in str(e):
                pass  # valid exception
            else:
                raise e

    def get_queue_attributes(self, queue_url, attribute_names=None):
        if attribute_names is None:
            attribute_names = ['All']
        try:
            return self.client.get_queue_attributes(QueueUrl=queue_url,
                                                    AttributeNames=attribute_names)
        except ClientError as e:
            if 'QueueDoesNotExistException' in str(e):
                pass  # valid exception
            else:
                raise e

    def update_queue(self, queue_url: str, delay_seconds=None,
                     maximum_message_size: int = None,
                     message_retention_period: int = None,
                     receive_message_wait_time_seconds: int = None,
                     policy: dict | str = None,
                     redrive_policy: dict | str = None,
                     visibility_timeout: int = None,
                     kms_master_key_id: str = None,
                     kms_data_key_reuse_period_seconds: int = None,
                     content_based_deduplication=None, tags: dict = None):
        errors = []
        attributes = dict()

        if delay_seconds is not None:
            if delay_seconds < 0 or delay_seconds > 900:
                errors.append(
                    'Delay seconds for SQS queue must be between '
                    '0 and 900 seconds'
                )
            else:
                attributes['DelaySeconds'] = str(delay_seconds)

        if maximum_message_size is not None:
            if maximum_message_size < 1024 or maximum_message_size > 262144:
                errors.append(
                    'Maximum message size must be between '
                    '1024 and 262144 bytes'
                )
            else:
                attributes['MaximumMessageSize'] = str(maximum_message_size)

        if message_retention_period is not None:
            if message_retention_period < 60 or message_retention_period > 1209600:
                errors.append(
                    'Message retention size must be between '
                    '60 and 1209600 seconds'
                )
            else:
                attributes['MessageRetentionPeriod'] = str(
                    message_retention_period)

        if receive_message_wait_time_seconds is not None:
            if receive_message_wait_time_seconds < 0 or receive_message_wait_time_seconds > 20:
                errors.append(
                    'Receive message wait time must be between '
                    '0 and 20 seconds'
                )
            else:
                attributes[
                    'ReceiveMessageWaitTimeSeconds'] = str(
                    receive_message_wait_time_seconds)

        if policy:
            if isinstance(policy, dict):
                policy = json.dumps(policy)
            attributes['Policy'] = policy

        if redrive_policy:
            if isinstance(policy, dict):
                redrive_policy = json.dumps(redrive_policy)
            attributes['RedrivePolicy'] = redrive_policy

        if visibility_timeout is not None:
            if visibility_timeout < 0 or visibility_timeout > 43200:
                errors.append(
                    'Visibility timeout must be between 0 and 43200 seconds'
                )
            else:
                attributes['VisibilityTimeout'] = str(visibility_timeout)

        if kms_master_key_id:
            attributes['KmsMasterKeyId'] = kms_master_key_id

        if kms_data_key_reuse_period_seconds is not None:
            if kms_data_key_reuse_period_seconds < 60 or kms_data_key_reuse_period_seconds > 86400:
                errors.append(
                    'KMS key reuse period must be between 60 and 86400 seconds'
                )
            else:
                attributes[
                    'KmsDataKeyReusePeriodSeconds'] = str(
                    kms_data_key_reuse_period_seconds)

        if content_based_deduplication:
            attributes[
                'ContentBasedDeduplication'] = str(content_based_deduplication)

        if errors:
            raise InvalidValueError(';\n'.join(errors))
        response = self.client.set_queue_attributes(
            QueueUrl=queue_url,
            Attributes=attributes
        )

        existing_tags = self.list_queue_tags(queue_url)
        tags_to_remove = [key for key in existing_tags if key not in tags]
        if tags_to_remove:
            self.untag_queue(queue_url, tags_to_remove)
        if tags:
            self.tag_queue(queue_url, tags)

        return response

    def list_queue_tags(self, queue_url):
        try:
            response = self.client.list_queue_tags(QueueUrl=queue_url)
            return response.get('Tags', dict())
        except ClientError as e:
            if 'AWS.SimpleQueueService.NonExistentQueue' in str(e):
                pass

    def untag_queue(self, queue_url: str, tags_to_remove: list):
        try:
            response = self.client.untag_queue(
                QueueUrl=queue_url,
                TagKeys=tags_to_remove
            )
            return response
        except ClientError as e:
            if 'AWS.SimpleQueueService.QueueDoesNotExist' in str(e):
                pass
            else:
                raise e

    def tag_queue(self, queue_url: str, tags: dict):
        try:
            response = self.client.tag_queue(
                QueueUrl=queue_url,
                Tags=tags
            )
            return response
        except ClientError as e:
            if 'AWS.SimpleQueueService.QueueDoesNotExist' in str(e):
                pass
            else:
                raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import boto3

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger(__name__)


@apply_methods_decorator(retry())
class ApplicationAutoscaling(object):
    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.region = region
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.client = boto3.client('application-autoscaling', region,
                                   aws_access_key_id=aws_access_key_id,
                                   aws_secret_access_key=aws_secret_access_key,
                                   aws_session_token=aws_session_token)
        _LOG.debug('Opened new Application autoscaling connection.')

    def register_target(self, service_namespace, resource_id,
                        scalable_dimension, min_capacity=None,
                        max_capacity=None, role_arn=None):
        params = dict(ServiceNamespace=service_namespace,
                      ResourceId=resource_id,
                      ScalableDimension=scalable_dimension)
        if min_capacity:
            params['MinCapacity'] = int(min_capacity)
        if max_capacity:
            params['MaxCapacity'] = int(max_capacity)
        if role_arn:
            params['RoleARN'] = role_arn
        self.client.register_scalable_target(**params)

    def put_step_scaling_policy(self, policy_name, service_namespace,
                                resource_id,
                                scalable_dimension,
                                adjustment_type=None,
                                metric_interval_lower_bound=None,
                                metric_interval_upper_bound=None,
                                scaling_adjustment=None,
                                min_adjustment_magnitude=None,
                                cooldown=None,
                                metric_aggregation_type=None):
        params = dict(PolicyName=policy_name,
                      ServiceNamespace=service_namespace,
                      ResourceId=resource_id,
                      ScalableDimension=scalable_dimension,
                      PolicyType='StepScaling')
        step_scaling_config_dict = dict()
        if adjustment_type:
            step_scaling_config_dict['AdjustmentType'] = adjustment_type
        if min_adjustment_magnitude:
            step_scaling_config_dict[
                'MinAdjustmentMagnitude'] = min_adjustment_magnitude
        if cooldown:
            step_scaling_config_dict['Cooldown'] = cooldown
        if metric_aggregation_type:
            step_scaling_config_dict[
                'MetricAggregationType'] = metric_aggregation_type
        step_adjustment_dict = dict()
        if metric_interval_lower_bound:
            step_adjustment_dict[
                'MetricIntervalLowerBound'] = metric_interval_lower_bound
        if metric_interval_upper_bound:
            step_adjustment_dict[
                'MetricIntervalUpperBound'] = metric_interval_upper_bound
        if scaling_adjustment:
            step_adjustment_dict['ScalingAdjustment'] = scaling_adjustment
        if step_adjustment_dict:
            step_scaling_config_dict['StepAdjustments'] = step_adjustment_dict
        if step_scaling_config_dict:
            params['StepScalingPolicyConfiguration'] = step_scaling_config_dict
        return self.client.put_scaling_policy(**params)

    def put_target_scaling_policy(self, policy_name, service_namespace,
                                  resource_id, scalable_dimension,
                                  target_value=None,
                                  predefined_metric_type=None,
                                  resource_label=None, metric_name=None,
                                  namespace=None, dimensions=None,
                                  statistic=None, unit=None,
                                  scale_out_cooldown=None,
                                  scale_in_cooldown=None):
        params = dict(PolicyName=policy_name,
                      ServiceNamespace=service_namespace,
                      ResourceId=resource_id,
                      ScalableDimension=scalable_dimension,
                      PolicyType='TargetTrackingScaling')
        target_scaling_config_dict = dict()
        if target_value:
            target_scaling_config_dict['TargetValue'] = target_value
        predefined_config_dict = dict()
        if predefined_metric_type:
            predefined_config_dict[
                'PredefinedMetricType'] = predefined_metric_type
        if resource_label:
            predefined_config_dict['ResourceLabel'] = resource_label
        if predefined_config_dict:
            target_scaling_config_dict[
                'PredefinedMetricSpecification'] = predefined_config_dict
        customized_config_dict = dict()
        if metric_name:
            customized_config_dict['MetricName'] = metric_name
        if namespace:
            customized_config_dict['Namespace'] = namespace
        if dimensions:
            customized_config_dict['Dimensions'] = dimensions
        if statistic:
            customized_config_dict['Statistic'] = statistic
        if unit:
            customized_config_dict['Unit'] = unit
        if customized_config_dict:
            params['CustomizedMetricSpecification'] = customized_config_dict
        if scale_out_cooldown:
            target_scaling_config_dict['ScaleOutCooldown'] = scale_out_cooldown
        if scale_in_cooldown:
            target_scaling_config_dict['ScaleInCooldown'] = scale_in_cooldown
        if target_scaling_config_dict:
            params[
                'TargetTrackingScalingPolicyConfiguration'] = target_scaling_config_dict
        return self.client.put_scaling_policy(**params)

    def deregister_scalable_target(self, service_namespace, resource_id,
                                   scalable_dimension):
        params = dict(ResourceId=resource_id,
                      ScalableDimension=scalable_dimension,
                      ServiceNamespace=service_namespace)
        return self.client.deregister_scalable_target(**params)

    def describe_scalable_targets(self, service_namespace, resources_ids=None,
                                  scalable_dimension=None):
        params = dict(ServiceNamespace=service_namespace)
        if resources_ids:
            params['ResourceIds'] = resources_ids
        if scalable_dimension:
            params['ScalableDimension'] = scalable_dimension
        targets = []
        response = self.client.describe_scalable_targets(**params)
        token = response.get('NextToken')
        targets.extend(response.get('ScalableTargets'))
        while token:
            params['NextToken'] = token
            response = self.client.describe_scalable_targets(**params)
            token = response.get('NextToken')
            targets.extend(response.get('ScalableTargets'))
        return targets

    def describe_scaling_policies(self, service_namespace, policy_names,
                                  resource_id=None, scalable_dimension=None):
        params = dict(ServiceNamespace=service_namespace)
        if policy_names:
            params['PolicyNames'] = policy_names
        if scalable_dimension:
            params['ScalableDimension'] = scalable_dimension
        if resource_id:
            params['ResourceId'] = resource_id
        targets = []
        response = self.client.describe_scaling_policies(**params)
        token = response.get('NextToken')
        targets.extend(response.get('ScalingPolicies'))
        while token:
            params['NextToken'] = token
            response = self.client.describe_scaling_policies(**params)
            token = response.get('NextToken')
            targets.extend(response.get('ScalingPolicies'))
        return targets



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from boto3 import client
from botocore.waiter import WaiterModel, create_waiter_with_client

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry
from syndicate.core.helper import dict_keys_to_camel_case

_LOG = get_logger(__name__)


@apply_methods_decorator(retry())
class BatchConnection(object):
    """ AWS Batch connection class. """

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('batch', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new Batch connection.')

    def create_compute_environment(self, compute_environment_name,
                                   compute_environment_type, state,
                                   service_role=None,
                                   compute_resources=None,
                                   tags=None):
        params = dict(
            computeEnvironmentName=compute_environment_name,
            type=compute_environment_type,
            state=state,
            serviceRole=service_role,
        )

        if compute_resources:
            compute_resources = dict_keys_to_camel_case(compute_resources)
            params['computeResources'] = compute_resources

        if tags:
            params['tags'] = tags

        return self.client.create_compute_environment(**params)

    def update_compute_environment(self, compute_environment, state=None,
                                   compute_resources=None, service_role=None):
        params = dict(computeEnvironment=compute_environment)

        if state:
            params['state'] = state

        if compute_resources:
            params['computeResources'] = dict_keys_to_camel_case(
                compute_resources)

        if service_role:
            params['serviceRole'] = service_role

        return self.client.update_compute_environment(**params)

    def describe_compute_environments(self, compute_environments):
        params = dict()
        if isinstance(compute_environments, str):
            params['computeEnvironments'] = [compute_environments]

        if isinstance(compute_environments, list):
            params['computeEnvironments'] = compute_environments
        return self.client.describe_compute_environments(**params)

    def delete_compute_environment(self, compute_environment):
        return self.client.delete_compute_environment(
            computeEnvironment=compute_environment)

    def create_job_queue(self, job_queue_name, state, priority,
                         compute_environment_order, tags=None):
        params = dict(
            jobQueueName=job_queue_name,
            state=state,
            priority=priority,
        )

        for index, item in enumerate(compute_environment_order):
            compute_environment_order[index] = dict_keys_to_camel_case(item)
        params['computeEnvironmentOrder'] = compute_environment_order

        if tags:
            params['tags'] = tags

        return self.client.create_job_queue(**params)

    def describe_job_queue(self, job_queues=None, max_results=None,
                           next_token=None):
        params = dict()
        if not job_queues:
            params['jobQueues'] = []
        if isinstance(job_queues, str):
            params['jobQueues'] = [job_queues]
        if isinstance(job_queues, list):
            params['jobQueues'] = job_queues

        if max_results:
            params['maxResults'] = max_results
        if next_token:
            params['nextToken'] = next_token

        return self.client.describe_job_queues(**params)

    def update_job_queue(self, job_queue, state=None, priority=None,
                         compute_environment_order=None):
        params = dict(jobQueue=job_queue)

        if state:
            params['state'] = state

        if priority:
            params['priority'] = priority

        if compute_environment_order:
            params['computeEnvironmentOrder'] = compute_environment_order

        return self.client.update_job_queue(**params)

    def delete_job_queue(self, job_queue):
        return self.client.delete_job_queue(
            jobQueue=job_queue
        )

    def register_job_definition(self, job_definition_name, job_definition_type,
                                parameters=None,
                                container_properties=None,
                                node_properties=None, retry_strategy=None,
                                propagate_tags=None,
                                timeout=None, tags=None,
                                platform_capabilities=None):

        params = dict(
            jobDefinitionName=job_definition_name,
            type=job_definition_type,
        )
        if parameters:
            params['parameters'] = dict_keys_to_camel_case(parameters)

        if container_properties:
            params['containerProperties'] = dict_keys_to_camel_case(
                container_properties)

        if node_properties:
            params['nodeProperties'] = dict_keys_to_camel_case(node_properties)

        if retry_strategy:
            params['retryStrategy'] = dict_keys_to_camel_case(retry_strategy)

        if propagate_tags is not None:
            params['propagateTags'] = propagate_tags
        if timeout:
            params['timeout'] = dict_keys_to_camel_case(timeout)
        if tags:
            params['tags'] = tags

        if platform_capabilities:
            params['platformCapabilities'] = platform_capabilities

        return self.client.register_job_definition(**params)

    def describe_job_definition(self, job_definition, max_results=None,
                                status=None):
        params = dict(jobDefinitionName=job_definition)
        if max_results:
            params['maxResults'] = max_results
        if status:
            params['status'] = status
        return self.client.describe_job_definitions(**params)

    def deregister_job_definition(self, job_definition_name):
        revisions = self._get_job_def_revisions(
            job_definition_name=job_definition_name)

        for revision in revisions:
            job_definition = '{0}:{1}'.format(job_definition_name, revision)
            self.client.deregister_job_definition(
                jobDefinition=job_definition
            )

    def get_compute_environment_waiter(self):
        waiter_id = 'ComputeEnvironmentWaiter'
        model = WaiterModel({
            'version': 2,
            'waiters': {
                waiter_id: {
                    'delay': 2,
                    'operation': 'DescribeComputeEnvironments',
                    'maxAttempts': 10,
                    'acceptors': [
                        {
                            'expected': 'VALID',
                            'matcher': 'pathAll',
                            'state': 'success',
                            'argument': 'computeEnvironments[].status'
                        },
                        {
                            'expected': 'INVALID',
                            'matcher': 'pathAny',
                            'state': 'failure',
                            'argument': 'computeEnvironments[].status'
                        }
                    ]
                }
            }
        })
        return create_waiter_with_client(waiter_id, model, self.client)

    def get_job_queue_waiter(self):
        waiter_id = 'JobQueueWaiter'
        model = WaiterModel({
            'version': 2,
            'waiters': {
                waiter_id: {
                    'delay': 1,
                    'operation': 'DescribeJobQueues',
                    'maxAttempts': 10,
                    'acceptors': [
                        {
                            'expected': 'VALID',
                            'matcher': 'pathAll',
                            'state': 'success',
                            'argument': 'jobQueues[].status'
                        },
                        {
                            'expected': 'INVALID',
                            'matcher': 'pathAny',
                            'state': 'failure',
                            'argument': 'jobQueues[].status'
                        }
                    ]
                }
            }
        })
        return create_waiter_with_client(waiter_id, model, self.client)

    def _get_job_def_revisions(self, job_definition_name):
        job_definition_data = self.describe_job_definition(
            job_definition=job_definition_name)
        revisions = []
        for job_def in job_definition_data['jobDefinitions']:
            if job_def.get('status') == 'ACTIVE':
                revisions.append(job_def.get('revision'))
        return revisions



"""
    Copyright 2024 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import sys
from pathlib import Path, PurePath

from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core import ProjectState, CONF_PATH
from syndicate.core.constants import APPSYNC_SCHEMA_DEFAULT_FILE_NAME, \
    APPSYNC_CONFIG_FILE_NAME
from syndicate.core.generators import _mkdir, _touch, _write_content_to_file
from syndicate.core.generators.contents import _generate_syncapp_config, \
    _generate_syncapp_default_schema
from syndicate.core.groups import RUNTIME_APPSYNC
from syndicate.core.project_state.project_state import BUILD_MAPPINGS


APPSYNC_FILES = [APPSYNC_SCHEMA_DEFAULT_FILE_NAME, APPSYNC_CONFIG_FILE_NAME]


_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


def generate_appsync(name, project_path, tags):
    path_to_project = Path(project_path)

    if not Path.exists(path_to_project):
        USER_LOG.info(f'Project "{project_path}" you '
                      f'have provided does not exist')
        return
    if not ProjectState.check_if_project_state_exists(CONF_PATH):
        USER_LOG.info(f'State file does not exist in {CONF_PATH}')
        return

    project_state = ProjectState(project_path=project_path)
    src_path = PurePath(project_path,
                        BUILD_MAPPINGS[RUNTIME_APPSYNC],
                        name).as_posix()
    if Path(src_path).exists():
        answer = _mkdir(
            path=src_path,
            fault_message=f'AppSync API with name \'{name}\' already exists.'
                          f'\nOverride? [y/n]')
        if not answer:
            USER_LOG.info(f'Creation of AppSync API \'{name}\' cancelled')
            sys.exit()
    else:
        _mkdir(src_path)

    default_schema_content = _generate_syncapp_default_schema()
    config_content = _generate_syncapp_config(
        name, APPSYNC_SCHEMA_DEFAULT_FILE_NAME, tags)

    for file_name in APPSYNC_FILES:
        path_to_file = PurePath(src_path, file_name).as_posix()
        _touch(path_to_file)

        if file_name == APPSYNC_SCHEMA_DEFAULT_FILE_NAME:
            _write_content_to_file(path_to_file, default_schema_content)

        if file_name == APPSYNC_CONFIG_FILE_NAME:
            _write_content_to_file(path_to_file, config_content)

    project_state.add_project_build_mapping(runtime=RUNTIME_APPSYNC)
    project_state.save()
    USER_LOG.info(f'AppSync API \'{name}\' added to the project')



"""
    Copyright 2024 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os

from syndicate.exceptions import ResourceMetadataError
from syndicate.core.constants import \
    EC2_LAUNCH_TEMPLATE_SUPPORTED_IMDS_VERSIONS


class LaunchTemplateValidator:
    def __init__(self, name, meta):
        self._name = name,
        self._meta = meta

    def validate(self):
        version_description = self._meta.get('version_description')
        if version_description is not None:
            if not isinstance(version_description, str):
                self._error(
                    "Unsupported type of the key 'version_description'. "
                    f"Supported type - 'str', current type - "
                    f"'{type(version_description).__name__}'")
        self._validate_lt_data(self._meta.get('launch_template_data'))

    def _error(self, message):
        raise ResourceMetadataError(message)

    def _validate_lt_data(self, lt_data):
        from syndicate.core import CONFIG
        if not lt_data:
            self._error('The launch_template_data is required.')

        if not isinstance(lt_data, dict):
            self._error(
                "Unsupported type of the key 'launch_template_data'. "
                f"Supported type - 'map', current type - "
                f"{type(lt_data).__name__}")

        imds_support = lt_data.get('imds_support')
        if imds_support is not None:
            if imds_support not in EC2_LAUNCH_TEMPLATE_SUPPORTED_IMDS_VERSIONS:
                self._error(f"The value of 'imds_support' - '{imds_support}' "
                            f"is invalid. Currently supported version - "
                            f"{EC2_LAUNCH_TEMPLATE_SUPPORTED_IMDS_VERSIONS}")

        user_data_file_path = lt_data.get('userdata_file')
        if user_data_file_path is not None:
            if not os.path.isabs(user_data_file_path):
                user_data_file_path = os.path.join(CONFIG.project_path,
                                                   user_data_file_path)
            if not os.path.isfile(user_data_file_path):
                self._error(f"There is no user data found by path "
                            f"'{user_data_file_path}'.")


def validate_launch_template(name, meta):
    LaunchTemplateValidator(name, meta).validate()



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger(__name__)


class BatchJobDefinitionResource(BaseResource):
    def __init__(self, batch_conn, iam_conn):
        self.batch_conn = batch_conn
        self.iam_conn = iam_conn

    def register_job_definition(self, args):
        return self.create_pool(self._register_job_definition_from_meta, args)

    def describe_job_definition(self, name, meta):
        response = self.batch_conn.describe_job_definition(job_definition=name,
                                                           max_results=1,
                                                           status='ACTIVE')
        try:
            arn = response['jobDefinitions'][-1]['jobDefinitionArn']
            return {arn: build_description_obj(response, name, meta)}
        except (KeyError, IndexError):
            _LOG.warn("Batch Job Definition %s not found", name)
            return {}

    def deregister_job_definition(self, args):
        return self.create_pool(self._deregister_job_definition, args)

    @unpack_kwargs
    def _register_job_definition_from_meta(self, name, meta):
        params = meta.copy()
        params['job_definition_name'] = name

        if 'resource_type' in params:
            del params['resource_type']

        container_properties = params.get('container_properties')
        if container_properties:
            job_role_arn = container_properties.get('job_role_arn')
            if job_role_arn:
                params['container_properties']['job_role_arn'] = self.iam_conn.check_if_role_exists(
                    role_name=job_role_arn
                )

            execution_role_arn = container_properties.get('execution_role_arn')
            if execution_role_arn:
                params['container_properties']['execution_role_arn'] = self.iam_conn.check_if_role_exists(
                    role_name=execution_role_arn
                )

        self.batch_conn.register_job_definition(**params)
        _LOG.info('Created Batch Job Definition %s.', name)
        return self.describe_job_definition(name, meta)

    @unpack_kwargs
    def _deregister_job_definition(self, arn, config):
        job_def_name = config['resource_name']
        self.batch_conn.deregister_job_definition(job_def_name)
        _LOG.info('Batch Job Definition %s was removed.', job_def_name)
        return {arn: config}

    @unpack_kwargs
    def _update_job_definition_from_meta(self, name, meta, context):
        """Updates batch job definition. If a user updates job definition,
        the previous active revision should be deregistered """
        response = self.batch_conn.describe_job_definition(job_definition=name,
                                                           max_results=1,
                                                           status='ACTIVE')
        previous_revision = response['jobDefinitions'][-1]['revision']

        self.batch_conn.client.deregister_job_definition(
            jobDefinition=f'{name}:{previous_revision}')
        BatchJobDefinitionResource._register_job_definition_from_meta(
            {
                'self': self,
                'name': name,
                'meta': meta
            }
        )

    def update_job_definition(self, args):
        self.create_pool(self._update_job_definition_from_meta, args)



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import FIREHOSE_TYPE

KINESIS_STREAM_FIREHOSE_TYPE = 'KinesisStreamAsSource'


class FirehoseGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = FIREHOSE_TYPE
    CONFIGURATION = {
        'stream_type': str,
        'kinesis_stream_source_configuration': None,
        's3_destination_configuration': dict,
        'tags': dict
    }

    def __init__(self, **kwargs):
        kwargs['s3_destination_configuration'] = {
            'role': kwargs.pop('destination_role'),
            'bucket': kwargs.pop('destination_bucket'),
            'buffering_hints': {
                'size_in_mbs': 5,
                'interval_in_seconds': 300
            },
            'compression_format': kwargs.pop('compression_format')
        }

        if kwargs['stream_type'] == KINESIS_STREAM_FIREHOSE_TYPE:
            kwargs['kinesis_stream_source_configuration'] = {
                'kinesis_stream_arn': kwargs.pop('kinesis_stream_arn'),
                'role': kwargs.pop('kinesis_stream_role')
            }

        super().__init__(**kwargs)



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from troposphere import Template

from syndicate.core.transform.build_meta_transformer import \
    BuildMetaTransformer
from .converter.cf_api_gateway_converter import CfApiGatewayConverter
from .converter.cf_batch_compenv_converter import \
    CfBatchComputeEnvironmentConverter
from .converter.cf_batch_job_defeinition import CfBatchJobDefinitionConverter
from .converter.cf_batch_job_queue_converter import CfBatchJobQueueConverter
from .converter.cf_cloudwatch_alarm_converter import CfCloudWatchAlarmConverter
from .converter.cf_cloudwatch_rule_converter import CfCloudWatchRuleConverter
from .converter.cf_cognito_converter import CfCognitoConverter
from .converter.cf_dynamodb_table_converter import CfDynamoDbTableConverter
from .converter.cf_iam_managed_policy_converter import \
    CfIamManagedPolicyConverter
from .converter.cf_iam_role_converter import CfIamRoleConverter
from .converter.cf_kinesis_stream_converter import CfKinesisStreamConverter
from .converter.cf_lambda_function_converter import CfLambdaFunctionConverter
from .converter.cf_lambda_layer_converter import CfLambdaLayerConverter
from .converter.cf_s3_converter import CfS3Converter
from .converter.cf_sns_converter import CfSnsConverter
from .converter.cf_sqs_converter import CfSqsConverter


class CloudFormationTransformer(BuildMetaTransformer):

    def __init__(self, bundle_name):
        super().__init__(bundle_name)
        self.template = Template()

    def output_file_name(self) -> str:
        return 'cloudformation_template.yaml'

    def _transform_iam_managed_policy(self, name, resource):
        self.convert_resources(name=name,
                               resource=resource,
                               converter_type=CfIamManagedPolicyConverter)

    def _transform_iam_role(self, name, resource):
        self.convert_resources(name=name,
                               resource=resource,
                               converter_type=CfIamRoleConverter)

    def _transform_lambda(self, name, resource):
        self.convert_resources(name=name,
                               resource=resource,
                               converter_type=CfLambdaFunctionConverter)

    def _transform_lambda_layer(self, name, resource):
        self.convert_resources(name=name,
                               resource=resource,
                               converter_type=CfLambdaLayerConverter)

    def _transform_dynamo_db_table(self, name, resource):
        self.convert_resources(name=name,
                               resource=resource,
                               converter_type=CfDynamoDbTableConverter)

    def _transform_s3_bucket(self, name, resource):
        self.convert_resources(name=name,
                               resource=resource,
                               converter_type=CfS3Converter)

    def _transform_cloud_watch_rule(self, name, resource):
        self.convert_resources(name=name,
                               resource=resource,
                               converter_type=CfCloudWatchRuleConverter)

    def _transform_api_gateway(self, name, resource):
        self.convert_resources(name=name,
                               resource=resource,
                               converter_type=CfApiGatewayConverter)

    def _transform_sns_topic(self, name, resource):
        self.convert_resources(name=name,
                               resource=resource,
                               converter_type=CfSnsConverter)

    def _transform_sns_application(self, name, resource):
        self.convert_resources(name=name,
                               resource=resource,
                               converter_type=CfSqsConverter)

    def _transform_cognito(self, name, resource):
        self.convert_resources(name=name,
                               resource=resource,
                               converter_type=CfCognitoConverter)

    def _transform_cloudwatch_alarm(self, name, resource):
        self.convert_resources(name=name,
                               resource=resource,
                               converter_type=CfCloudWatchAlarmConverter)

    def _transform_sqs_queue(self, name, resource):
        self.convert_resources(name=name,
                               resource=resource,
                               converter_type=CfSqsConverter)

    def _transform_kinesis_stream(self, name, resource):
        self.convert_resources(name=name,
                               resource=resource,
                               converter_type=CfKinesisStreamConverter)

    def _transform_batch_compenv(self, name, resource):
        self.convert_resources(
            name=name,
            resource=resource,
            converter_type=CfBatchComputeEnvironmentConverter)

    def _transform_batch_jobqueue(self, name, resource):
        self.convert_resources(
            name=name,
            resource=resource,
            converter_type=CfBatchJobQueueConverter)

    def _transform_batch_jobdef(self, name, resource):
        self.convert_resources(
            name=name,
            resource=resource,
            converter_type=CfBatchJobDefinitionConverter)

    def convert_resources(self, name, resource, converter_type):
        converter = converter_type(
            template=self.template,
            config=self.config,
            resources_provider=self.resources_provider)
        converter.convert(name=name, meta=resource)

    def _compose_template(self):
        return self.template.to_yaml()



from syndicate.exceptions import InvalidValueError
from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import DOCUMENTDB_CLUSTER_TYPE
from syndicate.commons.log_helper import get_logger, get_user_logger

_LOG = get_logger(
    'syndicate.core.generators.deployment_resources.documentdb_cluster')
USER_LOG = get_user_logger()


class DocumentDBClusterGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = DOCUMENTDB_CLUSTER_TYPE
    CONFIGURATION = {
        "vpc_security_group_ids": list,
        "availability_zones": list,
        "port": 27017,
        "master_username": None,
        "master_password": None,
        "tags": dict
    }

    def _generate_resource_configuration(self) -> dict:
        self._validate_master_password()
        return super()._generate_resource_configuration()

    def _validate_master_password(self):
        to_validate = self._dict.get('master_password')
        _LOG.info(f"Validating master password '{to_validate}'...")
        error = None
        if not 8 <= len(to_validate) <= 100:
            error = "master password must contain from 8 to 100 characters"
        elif any(char in to_validate for char in ('"@/')):
            error = 'master password cannot contain forward slash (/), ' \
                    'double quote (") or the "at" symbol (@)'
        if error:
            raise InvalidValueError(error)
        _LOG.info("Master password validation passed successfully")



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from botocore.exceptions import ClientError

from syndicate.exceptions import InvalidValueError, \
    ResourceNotFoundError
from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import (build_description_obj,
                                             validate_params)

CLOUDWATCH_ALARM_REQUIRED_PARAMS = ['metric_name', 'namespace', 'period',
                                    'threshold', 'evaluation_periods',
                                    'comparison_operator', 'statistic']

_LOG = get_logger(__name__)


class CloudWatchAlarmResource(BaseResource):

    def __init__(self, cw_conn, sns_conn, lambda_conn,
                 lambda_res, account_id) -> None:
        self.client = cw_conn
        self.sns_conn = sns_conn
        self.lambda_conn = lambda_conn
        self.lambda_res = lambda_res
        self.account_id = account_id

    def create_alarm(self, args):
        """ Create alarm in pool in sub processes.

        :type args: list
        """
        return self.create_pool(self._create_alarm_from_meta, args)

    def describe_alarm(self, name, meta):
        response = self.client.describe_alarms([name])
        if not response:
            return {}
        arn = response[0]['AlarmArn']
        return {
            arn: build_description_obj(response, name, meta)
        }

    @unpack_kwargs
    def _create_alarm_from_meta(self, name, meta):
        """ Create alarm resource in AWS Cloud via meta description.

        :type name: str
        :type meta: dict
        """
        validate_params(name, meta, CLOUDWATCH_ALARM_REQUIRED_PARAMS)

        if self.client.is_alarm_exists(name):
            _LOG.warning(f'{name} alarm exists.')
            return self.describe_alarm(name, meta)

        params = dict(alarm_name=name, metric_name=meta['metric_name'],
                      namespace=meta['namespace'], period=meta['period'],
                      evaluation_periods=meta['evaluation_periods'],
                      threshold=meta['threshold'], statistic=meta['statistic'],
                      comparison_operator=meta['comparison_operator'],
                      description=meta.get('description'),
                      dimensions=meta.get('dimensions'),
                      datapoints=meta.get('datapoints'), alarm_actions=[],
                      evaluate_low_sample_count_percentile=meta.get(
                          'evaluate_low_sample_count_percentile'),
                      tags=meta.get('tags'))

        if sns_topics := meta.get('sns_topics'):
            for each in sns_topics:
                if arn := self.sns_conn.get_topic_arn(each):
                    params['alarm_actions'].append(arn)
        if lambdas := meta.get('lambdas'):
            for each in lambdas:
                if arn := self._validate_lambda_resource(each):
                    params['alarm_actions'].append(arn)
        if response_plans := meta.get('ssm_response_plan'):
            for each in response_plans:
                params['alarm_actions'].append(
                    f'arn:aws:ssm-incidents::{self.account_id}:responseplan/{each}')

        self.client.put_metric_alarm(**params)
        _LOG.info(f'Created alarm {name}.')
        return self.describe_alarm(name, meta)

    def remove_alarms(self, args):
        return self.create_pool(self._remove_alarms, args)

    @unpack_kwargs
    def _remove_alarms(self, arn, config):
        alarm_name = config['resource_name']
        try:
            self.client.remove_alarms(alarm_names=[alarm_name],
                                      log_not_found_error=False)
            _LOG.info(f'Alarm {alarm_name} was removed.')

        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                _LOG.warn(f'Alarm {alarm_name} not found.')
            else:
                raise e
            described_alarms = self.client.alarm_list(alarm_names=[alarm_name])
            if described_alarms and any(alarm['AlarmName'] == alarm_name
                                        for alarm in described_alarms):
                _LOG.warn(f'Alarm {alarm_name} was found despite the '
                          f'`ResourceNotFoundException` error.')
                raise e

        return {arn: config}

    def _validate_lambda_resource(self, resource: str) -> str | None:
        from syndicate.core import CONFIG
        qualifier = None
        if ':' in resource:
            if resource.count(':') > 1:
                raise InvalidValueError(
                    f'Invalid lambda qualifier \'{resource}\''
                )

            resource_name, qualifier = resource.split(':')
            resource = f'{CONFIG.resources_prefix}{resource_name}' \
                       f'{CONFIG.resources_suffix}'

        arn = self.lambda_res.build_lambda_arn(resource)
        if self.lambda_conn.get_function(arn, qualifier):
            return arn
        else:
            raise ResourceNotFoundError(
                f'Cannot find lambda with arn \'{arn}\' '
                f'and qualifier \'{qualifier}\''
            )



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from syndicate.core.constants import \
    API_GATEWAY_TYPE, WEB_SOCKET_API_GATEWAY_TYPE, CLOUD_WATCH_ALARM_TYPE, \
    CLOUD_WATCH_RULE_TYPE, EVENT_BRIDGE_RULE_TYPE, DYNAMO_TABLE_TYPE, \
    EBS_TYPE, EC2_INSTANCE_TYPE, IAM_POLICY, IAM_ROLE, KINESIS_STREAM_TYPE, \
    LAMBDA_TYPE, S3_BUCKET_TYPE, SNS_PLATFORM_APPLICATION_TYPE, \
    SNS_TOPIC_TYPE, SQS_QUEUE_TYPE, STATE_ACTIVITY_TYPE, STEP_FUNCTION_TYPE, \
    LAMBDA_LAYER_TYPE, BATCH_COMPENV_TYPE, BATCH_JOBQUEUE_TYPE, \
    BATCH_JOBDEF_TYPE, COGNITO_USER_POOL_TYPE, COGNITO_FEDERATED_POOL_TYPE, \
    DOCUMENTDB_CLUSTER_TYPE, DOCUMENTDB_INSTANCE_TYPE, DAX_CLUSTER_TYPE, \
    FIREHOSE_TYPE, EVENT_BRIDGE_SCHEDULE_TYPE, API_GATEWAY_OAS_V3_TYPE, \
    SWAGGER_UI_TYPE, EC2_LAUNCH_TEMPLATE_TYPE, APPSYNC_TYPE, \
    RDS_DB_CLUSTER_TYPE, RDS_DB_INSTANCE_TYPE


class ProcessorFacade:

    def __init__(self, resources_provider) -> None:
        self.resources_provider = resources_provider

    def create_handlers(self):
        return {
            IAM_POLICY:
                self.resources_provider.iam().create_policies,
            IAM_ROLE:
                self.resources_provider.iam().create_roles,
            LAMBDA_TYPE:
                self.resources_provider.lambda_resource().create_lambda,
            LAMBDA_LAYER_TYPE:
                self.resources_provider.lambda_resource().create_lambda_layer,
            DYNAMO_TABLE_TYPE:
                self.resources_provider.dynamodb().create_tables,
            CLOUD_WATCH_RULE_TYPE:
                self.resources_provider.cw().create_cloud_watch_rule,
            EVENT_BRIDGE_RULE_TYPE:
                self.resources_provider.cw().create_cloud_watch_rule,
            S3_BUCKET_TYPE:
                self.resources_provider.s3().create_s3_bucket,
            API_GATEWAY_TYPE:
                self.resources_provider.api_gw().create_api_gateway,
            API_GATEWAY_OAS_V3_TYPE:
                self.resources_provider.api_gw().create_api_gateway_openapi,
            WEB_SOCKET_API_GATEWAY_TYPE:
                self.resources_provider.api_gw().create_web_socket_api_gateway,
            COGNITO_USER_POOL_TYPE:
                self.resources_provider.cognito_user_pool()
                    .create_cognito_user_pool,
            COGNITO_FEDERATED_POOL_TYPE:
                self.resources_provider.cognito_identity()
                    .create_cognito_identity_pool,
            SNS_TOPIC_TYPE:
                self.resources_provider.sns().create_sns_topic,
            SNS_PLATFORM_APPLICATION_TYPE:
                self.resources_provider.sns().create_sns_application,
            SQS_QUEUE_TYPE:
                self.resources_provider.sqs().create_sqs_queue,
            CLOUD_WATCH_ALARM_TYPE:
                self.resources_provider.cw_alarm().create_alarm,
            EBS_TYPE:
                self.resources_provider.ebs().create_ebs,
            STEP_FUNCTION_TYPE:
                self.resources_provider.step_functions().create_state_machine,
            STATE_ACTIVITY_TYPE:
                self.resources_provider.step_functions().create_activities,
            KINESIS_STREAM_TYPE:
                self.resources_provider.kinesis().create_kinesis_stream,
            EC2_INSTANCE_TYPE:
                self.resources_provider.ec2().create_ec2,
            EC2_LAUNCH_TEMPLATE_TYPE:
                self.resources_provider.ec2().create_launch_template,
            BATCH_COMPENV_TYPE:
                self.resources_provider.batch_compenv()
                    .create_compute_environment,
            BATCH_JOBQUEUE_TYPE:
                self.resources_provider.batch_jobqueue().create_job_queue,
            BATCH_JOBDEF_TYPE:
                self.resources_provider.batch_jobdef().register_job_definition,
            DOCUMENTDB_CLUSTER_TYPE:
                self.resources_provider.documentdb_cluster().create_db_cluster,
            DOCUMENTDB_INSTANCE_TYPE:
                self.resources_provider.documentdb_instance()
                    .create_db_instance,
            DAX_CLUSTER_TYPE:
                self.resources_provider.dax_cluster().create_cluster,
            FIREHOSE_TYPE:
                self.resources_provider.firehose().create_stream,
            EVENT_BRIDGE_SCHEDULE_TYPE:
                self.resources_provider.eventbridge_scheduler().create_schedule,
            SWAGGER_UI_TYPE: self.resources_provider.swagger_ui().
            create_update_swagger_ui,
            APPSYNC_TYPE:
                self.resources_provider.appsync().create_graphql_api,
            RDS_DB_CLUSTER_TYPE:
                self.resources_provider.rds_db_cluster().create_db_cluster,
            RDS_DB_INSTANCE_TYPE:
                self.resources_provider.rds_db_instance().create_db_instance
        }

    def describe_handlers(self):
        return {
            IAM_POLICY:
                self.resources_provider.iam().describe_policy,
            IAM_ROLE:
                self.resources_provider.iam().describe_role,
            LAMBDA_TYPE:
                self.resources_provider.lambda_resource().describe_lambda,
            DYNAMO_TABLE_TYPE:
                self.resources_provider.dynamodb().describe_table,
            CLOUD_WATCH_RULE_TYPE:
                self.resources_provider.cw().describe_rule_from_meta,
            EVENT_BRIDGE_RULE_TYPE:
                self.resources_provider.cw().describe_rule_from_meta,
            S3_BUCKET_TYPE:
                self.resources_provider.s3().describe_bucket,
            API_GATEWAY_TYPE:
                self.resources_provider.api_gw().describe_api_resources,
            API_GATEWAY_OAS_V3_TYPE:
                self.resources_provider.api_gw().describe_api_resources,
            WEB_SOCKET_API_GATEWAY_TYPE:
                self.resources_provider.api_gw().describe_v2_api_gateway,
            COGNITO_USER_POOL_TYPE:
                self.resources_provider.cognito_user_pool()
                    .describe_user_pool,
            COGNITO_FEDERATED_POOL_TYPE:
                self.resources_provider.cognito_identity().describe_cognito_pool,
            SNS_TOPIC_TYPE:
                self.resources_provider.sns().describe_sns_from_meta,
            SNS_PLATFORM_APPLICATION_TYPE:
                self.resources_provider.sns()
                    .describe_sns_application_from_meta,
            SQS_QUEUE_TYPE:
                self.resources_provider.sqs().describe_queue_from_meta,
            CLOUD_WATCH_ALARM_TYPE:
                self.resources_provider.cw_alarm().describe_alarm,
            EBS_TYPE:
                self.resources_provider.ebs().describe_ebs,
            STEP_FUNCTION_TYPE:
                self.resources_provider.step_functions()
                    .describe_step_function,
            STATE_ACTIVITY_TYPE:
                self.resources_provider.step_functions().describe_activity,
            KINESIS_STREAM_TYPE:
                self.resources_provider.kinesis().describe_kinesis_stream,
            BATCH_COMPENV_TYPE:
                self.resources_provider.batch_compenv()
                    .describe_compute_environment,
            BATCH_JOBQUEUE_TYPE:
                self.resources_provider.batch_jobqueue().describe_job_queue,
            BATCH_JOBDEF_TYPE:
                self.resources_provider.batch_jobdef().describe_job_definition,
            DOCUMENTDB_CLUSTER_TYPE:
                self.resources_provider.documentdb_cluster()
                    .describe_documentdb_cluster,
            DOCUMENTDB_INSTANCE_TYPE:
                self.resources_provider.documentdb_instance()
                    .describe_documentdb_instance,
            DAX_CLUSTER_TYPE:
                self.resources_provider.dax_cluster().describe_cluster,
            FIREHOSE_TYPE:
                self.resources_provider.firehose().describe_stream,
            EVENT_BRIDGE_SCHEDULE_TYPE:
                self.resources_provider.eventbridge_scheduler().describe_schedule,
            SWAGGER_UI_TYPE: self.resources_provider.swagger_ui().
            describe_swagger_ui,
            EC2_LAUNCH_TEMPLATE_TYPE:
                self.resources_provider.ec2().describe_launch_template,
            LAMBDA_LAYER_TYPE:
                self.resources_provider.lambda_resource().describe_lambda_layer,
            APPSYNC_TYPE:
                self.resources_provider.appsync().describe_graphql_api,
            RDS_DB_CLUSTER_TYPE:
                self.resources_provider.rds_db_cluster().describe_db_cluster,
            RDS_DB_INSTANCE_TYPE:
                self.resources_provider.rds_db_instance().describe_db_instance
        }

    def remove_handlers(self):
        return {
            CLOUD_WATCH_ALARM_TYPE:
                self.resources_provider.cw_alarm().remove_alarms,
            API_GATEWAY_TYPE:
                self.resources_provider.api_gw().remove_api_gateways,
            API_GATEWAY_OAS_V3_TYPE:
                self.resources_provider.api_gw().remove_api_gateways,
            WEB_SOCKET_API_GATEWAY_TYPE:
                self.resources_provider.api_gw().remove_v2_api_gateway,
            CLOUD_WATCH_RULE_TYPE:
                self.resources_provider.cw().remove_cloud_watch_rules,
            EVENT_BRIDGE_RULE_TYPE:
                self.resources_provider.cw().remove_cloud_watch_rules,
            COGNITO_USER_POOL_TYPE:
                self.resources_provider.cognito_user_pool()
                    .remove_cognito_user_pools,
            COGNITO_FEDERATED_POOL_TYPE:
                self.resources_provider.cognito_identity()
                    .remove_cognito_identity_pools,
            DYNAMO_TABLE_TYPE:
                self.resources_provider.dynamodb().remove_dynamodb_tables,
            EBS_TYPE:
                self.resources_provider.ebs().remove_ebs_apps,
            EC2_INSTANCE_TYPE:
                self.resources_provider.ec2().remove_ec2_instances,
            EC2_LAUNCH_TEMPLATE_TYPE:
                self.resources_provider.ec2().remove_launch_templates,
            IAM_ROLE:
                self.resources_provider.iam().remove_roles,
            IAM_POLICY:
                self.resources_provider.iam().remove_policies,
            KINESIS_STREAM_TYPE:
                self.resources_provider.kinesis().remove_kinesis_streams,
            LAMBDA_LAYER_TYPE:
                self.resources_provider.lambda_resource().remove_lambda_layers,
            LAMBDA_TYPE:
                self.resources_provider.lambda_resource().remove_lambdas,
            S3_BUCKET_TYPE:
                self.resources_provider.s3().remove_buckets,
            SNS_TOPIC_TYPE:
                self.resources_provider.sns().remove_sns_topics,
            SNS_PLATFORM_APPLICATION_TYPE:
                self.resources_provider.sns().remove_sns_application,
            SQS_QUEUE_TYPE:
                self.resources_provider.sqs().remove_queues,
            STEP_FUNCTION_TYPE:
                self.resources_provider.step_functions().remove_state_machines,
            STATE_ACTIVITY_TYPE:
                self.resources_provider.step_functions().remove_activities,
            BATCH_COMPENV_TYPE:
                self.resources_provider.batch_compenv()
                    .remove_compute_environment,
            BATCH_JOBQUEUE_TYPE:
                self.resources_provider.batch_jobqueue().remove_job_queue,
            BATCH_JOBDEF_TYPE:
                self.resources_provider.batch_jobdef()
                    .deregister_job_definition,
            DOCUMENTDB_CLUSTER_TYPE:
                self.resources_provider.documentdb_cluster().remove_db_cluster,
            DOCUMENTDB_INSTANCE_TYPE:
                self.resources_provider.documentdb_instance()
                    .remove_db_instance,
            DAX_CLUSTER_TYPE:
                self.resources_provider.dax_cluster().remove_cluster,
            FIREHOSE_TYPE:
                self.resources_provider.firehose().delete_streams,
            EVENT_BRIDGE_SCHEDULE_TYPE:
                self.resources_provider.eventbridge_scheduler().remove_schedule,
            SWAGGER_UI_TYPE:
                self.resources_provider.swagger_ui().remove_swagger_ui,
            APPSYNC_TYPE:
                self.resources_provider.appsync().remove_graphql_api,
            RDS_DB_CLUSTER_TYPE:
                self.resources_provider.rds_db_cluster().remove_db_cluster,
            RDS_DB_INSTANCE_TYPE:
                self.resources_provider.rds_db_instance().remove_db_instance
        }

    def update_handlers(self):
        return {
            IAM_ROLE:
                self.resources_provider.iam().update_iam_role,
            IAM_POLICY:
                self.resources_provider.iam().update_iam_policy,
            SQS_QUEUE_TYPE:
                self.resources_provider.sqs().update_sqs_queue,
            LAMBDA_TYPE:
                self.resources_provider.lambda_resource().update_lambda,
            LAMBDA_LAYER_TYPE:
                self.resources_provider.lambda_resource().update_lambda_layer,
            API_GATEWAY_OAS_V3_TYPE:
                self.resources_provider.api_gw().update_api_gateway_openapi,
            BATCH_JOBDEF_TYPE:
                self.resources_provider.batch_jobdef().update_job_definition,
            BATCH_COMPENV_TYPE:
                self.resources_provider.batch_compenv()
                    .update_compute_environment,
            EVENT_BRIDGE_SCHEDULE_TYPE:
                self.resources_provider.eventbridge_scheduler()
                    .update_schedule,
            DYNAMO_TABLE_TYPE:
                self.resources_provider.dynamodb().update_tables,
            SWAGGER_UI_TYPE: self.resources_provider.swagger_ui().
            create_update_swagger_ui,
            EC2_LAUNCH_TEMPLATE_TYPE:
                self.resources_provider.ec2().update_launch_template,
            APPSYNC_TYPE:
                self.resources_provider.appsync().update_graphql_api,
            RDS_DB_CLUSTER_TYPE:
                self.resources_provider.rds_db_cluster().update_db_cluster,
            RDS_DB_INSTANCE_TYPE:
                self.resources_provider.rds_db_instance().update_db_instance
        }

    def resource_configuration_processor(self):
        return {
            API_GATEWAY_TYPE:
                self.resources_provider.api_gw().api_gateway_update_processor
        }

    def resource_identifier(self):
        return {
            API_GATEWAY_TYPE:
                self.resources_provider.api_gw().api_resource_identifier,
            COGNITO_USER_POOL_TYPE:
                self.resources_provider.cognito_user_pool()
                    .cognito_resource_identifier,
            COGNITO_FEDERATED_POOL_TYPE:
                self.resources_provider.cognito_identity()
                    .cognito_resource_identifier
        }

    def mapping_applier(self):
        return {
            IAM_ROLE: self.resources_provider.iam().apply_trusted_to_role,
            IAM_POLICY: self.resources_provider.iam().apply_policy_content
        }

    def compare_meta_handlers(self):
        return {
            DYNAMO_TABLE_TYPE: self.resources_provider.dynamodb().compare_meta,
        }



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import sys
import traceback
import threading
from functools import wraps
from pathlib import PurePath

import click
from click import BadParameter

from syndicate.exceptions import SyndicateBaseError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import (ABORTED_RETURN_CODE, OK_RETURN_CODE,
                                      FAILED_RETURN_CODE)

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()

lock = threading.Lock()


def check_deploy_name_for_duplicates(func):
    """
    Checks whether output file with specified name already exists.
    Everywhere this decorator is used the following
    :param func:
    :return:
    """

    @wraps(func)
    def real_wrapper(*args, **kwargs):
        from syndicate.core import CONN
        from syndicate.core import CONFIG
        deploy_name = kwargs.get('deploy_name')
        bundle_name = kwargs.get('bundle_name')
        replace_output = kwargs.get('replace_output')
        if deploy_name and bundle_name and not replace_output:
            output_file_name = f'{bundle_name}/outputs/{deploy_name}.json'

            key_compound = PurePath(CONFIG.deploy_target_bucket_key_compound,
                                    output_file_name).as_posix()
            exists = CONN.s3().is_file_exists(CONFIG.deploy_target_bucket,
                                              key=key_compound)
            if exists:
                msg = f'Output file already exists with name ' \
                      f'{output_file_name}. If it should be replaced with ' \
                      f'new one, use --replace-output flag.'
                USER_LOG.error(msg)
                return ABORTED_RETURN_CODE
        return func(*args, **kwargs)

    return real_wrapper


def check_deploy_bucket_exists(func):
    @wraps(func)
    def real_wrapper(*args, **kwargs):
        from syndicate.core import CONN
        from syndicate.core import CONFIG
        if not CONN.s3().is_bucket_exists(CONFIG.deploy_target_bucket):
            USER_LOG.error(
                'Cannot execute command: deploy target bucket does not exist. '
                'Please create it before executing commands that require '
                'files to be uploaded to the bucket.')
            return ABORTED_RETURN_CODE
        return func(*args, **kwargs)

    return real_wrapper


def threading_lock(func):
    """ Synchronize access to a function with a threading lock
    to avoid race condition."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        with lock:
            _LOG.info('Lock acquired')
            result = func(*args, **kwargs)
        _LOG.info('Lock released')
        return result

    return wrapper


def check_bundle_deploy_names_for_existence(check_deploy_existence=False):
    def internal_check(func):
        @wraps(func)
        def real_wrapper(*args, **kwargs):
            from syndicate.core.build.bundle_processor import if_bundle_exist
            from syndicate.core.build.deployment_processor import is_deploy_exist

            deploy_name = kwargs.get('deploy_name')
            bundle_name = kwargs.get('bundle_name')
            if not bundle_name:
                USER_LOG.error(f'The bundle name is undefined or invalid, '
                               f'please verify it and try again.')
                return ABORTED_RETURN_CODE
            if not deploy_name:
                USER_LOG.error(f'The deploy name is undefined or invalid, '
                               f'please verify it and try again.')
                return ABORTED_RETURN_CODE
            if not if_bundle_exist(bundle_name=bundle_name):
                USER_LOG.error(
                    f'The bundle name \'{bundle_name}\' does not exist '
                    f'in deploy bucket. Please verify the bundle name '
                    f'and try again.')
                return ABORTED_RETURN_CODE
            if check_deploy_existence and not is_deploy_exist(
                    bundle_name=bundle_name, deploy_name=deploy_name):
                USER_LOG.error(
                    f'The deploy name \'{deploy_name}\' is invalid. '
                    f'Please verify the deploy name and try again.')
                return ABORTED_RETURN_CODE
            return func(*args, **kwargs)
        return real_wrapper
    return internal_check


def return_code_manager(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return_code = func(*args, **kwargs)
        except Exception as e:
            if isinstance(e, BadParameter):
                message = f"{e.__class__.__name__} {e.message}"
            elif isinstance(e, SyndicateBaseError):
                message = f"{e.__class__.__name__} occurred: {str(e)}"
            else:
                message = (f'An unexpected error occurred: '
                           f'{e.__class__.__name__} {str(e)}')

            USER_LOG.error(message)
            _LOG.exception(traceback.format_exc())

            sys.exit(FAILED_RETURN_CODE)
        if return_code is not None and return_code != OK_RETURN_CODE:
            sys.exit(return_code)

        return return_code
    return wrapper


def tags_to_context(func):
    """
    Decorator to load resource tags and output from the latest deploy
    to click context
    """

    @click.pass_context
    @wraps(func)
    def wrapper(ctx, *args, **kwargs):
        from syndicate.core import PROJECT_STATE, RESOURCES_PROVIDER
        from syndicate.core.build.bundle_processor import load_deploy_output
        if PROJECT_STATE is None or not PROJECT_STATE.latest_deploy:
            USER_LOG.error(
                'Project state is not found/not initialized. '
                'Please check your configuration.'
            )
            sys.exit(FAILED_RETURN_CODE)

        deploy_name = PROJECT_STATE.latest_deploy.get('deploy_name')
        bundle_name = PROJECT_STATE.latest_deploy.get('bundle_name')
        output = load_deploy_output(bundle_name, deploy_name)

        ctx.ensure_object(dict)
        ctx.obj['output'] = output
        ctx.obj['tags_resource'] = RESOURCES_PROVIDER.tags_api()

        return_code = func(*args, **kwargs)
        return return_code
    return wrapper



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import COGNITO_FEDERATED_POOL_TYPE


class CognitoFederatedPoolGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = COGNITO_FEDERATED_POOL_TYPE
    CONFIGURATION = {
        "auth_role": None,
        "unauth_role": None,
        "open_id_providers": list,
        "provider_name": None,
        "tags": dict
    }



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import os
import shutil
from concurrent.futures import ThreadPoolExecutor
from pathlib import PurePath
from botocore.exceptions import ClientError

from syndicate.exceptions import ProjectStateError, ConfigurationError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.connection import S3Connection
from syndicate.core.build.helper import _json_serial, resolve_bundle_directory, \
    resolve_all_bundles_directory, assert_bundle_bucket_exists
from syndicate.core.build.meta_processor import validate_deployment_packages, \
    preprocess_tags
from syndicate.core.constants import (ARTIFACTS_FOLDER, BUILD_META_FILE_NAME,
                                      DEFAULT_SEP)
from syndicate.core.helper import build_path, unpack_kwargs

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


def _build_output_key(bundle_name, deploy_name, is_regular_output):
    return '{0}/outputs/{1}{2}.json'.format(
        bundle_name, deploy_name, '' if is_regular_output else '_failed')


def _backup_deploy_output(filename, output):
    _LOG.info('Wrote file to {0}'.format(filename))
    with open(filename, 'a+') as backup_file:
        backup_file.write(output)
        backup_file.close()


def create_deploy_output(
        bundle_name: str,
        deploy_name: str,
        output: dict,
        success: bool,
        replace_output: bool = False,
) -> None:
    from syndicate.core import CONFIG, CONN
    _LOG.debug('Going to preprocess resources tags in output')
    preprocess_tags(output)
    output_str = json.dumps(output, default=_json_serial)
    key = _build_output_key(bundle_name=bundle_name,
                            deploy_name=deploy_name,
                            is_regular_output=success)
    key_compound = \
        PurePath(CONFIG.deploy_target_bucket_key_compound, key).as_posix()
    if CONN.s3().is_file_exists(CONFIG.deploy_target_bucket, key_compound) \
            and not replace_output:
        _LOG.warning(f'Output file for deploy {deploy_name} already exists')
    else:
        CONN.s3().put_object(output_str, key_compound,
                             CONFIG.deploy_target_bucket,
                             'application/json')
        _LOG.info(
            f"Output file with name {key} has been "
            f"{'replaced' if replace_output else 'created'}"
        )


def remove_deploy_output(
        bundle_name: str,
        deploy_name: str,
) -> None:
    from syndicate.core import CONFIG, CONN
    key = _build_output_key(bundle_name=bundle_name,
                            deploy_name=deploy_name,
                            is_regular_output=True)
    key_compound = \
        PurePath(CONFIG.deploy_target_bucket_key_compound, key).as_posix()
    if CONN.s3().is_file_exists(CONFIG.deploy_target_bucket, key_compound):
        CONN.s3().remove_object(CONFIG.deploy_target_bucket, key_compound)
    else:
        _LOG.warning(f'Output file for deploy {deploy_name} does not exist')


def remove_failed_deploy_output(
        bundle_name: str,
        deploy_name: str,
) -> None:
    from syndicate.core import CONFIG, CONN
    key = _build_output_key(bundle_name=bundle_name,
                            deploy_name=deploy_name,
                            is_regular_output=False)
    key_compound = \
        PurePath(CONFIG.deploy_target_bucket_key_compound, key).as_posix()
    if CONN.s3().is_file_exists(CONFIG.deploy_target_bucket, key_compound):
        _LOG.debug(
            f"Going to remove failed output '{key_compound}' from the bucket "
            f"'{CONFIG.deploy_target_bucket}'"
        )
        CONN.s3().remove_object(CONFIG.deploy_target_bucket, key_compound)
    else:
        _LOG.warning(
            f'Failed output file for deploy {deploy_name} does not exist'
        )


def load_deploy_output(
        bundle_name: str,
        deploy_name: str,
        failsafe: bool = False,
) -> dict | bool:
    """
    :param bundle_name:
    :param deploy_name:
    :param failsafe: False - raise error if no deploy output;
                     True - do not raise an error, return False
    """
    from syndicate.core import CONFIG, CONN
    key = _build_output_key(bundle_name=bundle_name,
                            deploy_name=deploy_name,
                            is_regular_output=True)
    key_compound = \
        PurePath(CONFIG.deploy_target_bucket_key_compound, key).as_posix()
    if CONN.s3().is_file_exists(CONFIG.deploy_target_bucket, key_compound):
        output_file = \
            CONN.s3().load_file_body(CONFIG.deploy_target_bucket, key_compound)
        return json.loads(output_file)
    else:
        if failsafe:
            _LOG.warning(
                f'Deploy name {deploy_name} does not exist. '
                f'Failsafe status - {failsafe}'
            )
            return False
        raise ProjectStateError(
            f"Cannot find output file for the deploy name '{deploy_name}'."
        )


def load_failed_deploy_output(bundle_name, deploy_name,
                              failsafe: bool = False):
    """
    :param bundle_name:
    :param deploy_name:
    :param failsafe: False - raise error if no deploy output;
                     True - do not raise an error, return False
    """
    from syndicate.core import CONFIG, CONN
    key = _build_output_key(bundle_name=bundle_name,
                            deploy_name=deploy_name,
                            is_regular_output=False)
    key_compound = PurePath(CONFIG.deploy_target_bucket_key_compound,
                            key).as_posix()
    if CONN.s3().is_file_exists(CONFIG.deploy_target_bucket,
                                key_compound):
        output_file = CONN.s3().load_file_body(
            CONFIG.deploy_target_bucket,
            key_compound)
        return json.loads(output_file)
    else:
        if failsafe:
            _LOG.warn(f'Deploy name {deploy_name} does not exist. '
                      f'Failsafe status - {failsafe}.')
            return False
        raise ProjectStateError(
            f"Cannot find output file for the deploy name '{deploy_name}'."
        )


def load_latest_deploy_output(failsafe: bool = False):
    from syndicate.core import PROJECT_STATE
    if not PROJECT_STATE.latest_deploy:
        return None, {}
    deploy_name = PROJECT_STATE.latest_deployed_deploy_name
    bundle_name = PROJECT_STATE.latest_deployed_bundle_name
    latest_deploy_status = PROJECT_STATE.latest_deploy.get(
        'is_succeeded', True)

    if latest_deploy_status is True:
        return True, load_deploy_output(bundle_name, deploy_name,
                                        failsafe=failsafe)
    elif latest_deploy_status is False:
        return False, load_failed_deploy_output(bundle_name, deploy_name,
                                                failsafe=failsafe)
    else:
        raise ProjectStateError(
            "The latest deployments' status can't be resolved because of "
            "unexpected status. Please check the parameter 'is_succeeded' "
            "value in the 'latest_deploy' section of the syndicate state "
            "file. Expected value is 'true' or 'false'")


def load_meta_resources(bundle_name):
    from syndicate.core import CONFIG, CONN
    key = build_path(bundle_name, BUILD_META_FILE_NAME)
    key_compound = PurePath(CONFIG.deploy_target_bucket_key_compound,
                            key).as_posix()
    meta_file = CONN.s3().load_file_body(
        CONFIG.deploy_target_bucket, key_compound)
    return json.loads(meta_file)


def if_bundle_exist(bundle_name):
    from syndicate.core import CONFIG, CONN
    assert_bundle_bucket_exists()
    bundle_folder = bundle_name + DEFAULT_SEP
    key_compound = PurePath(CONFIG.deploy_target_bucket_key_compound,
                            bundle_folder).as_posix()
    return CONN.s3().get_keys_by_prefix(
        CONFIG.deploy_target_bucket,
        key_compound)


def if_bundle_exist_locally(bundle_name):
    bundle_dir = resolve_bundle_directory(bundle_name=bundle_name)
    normalized_bundle_dir = os.path.normpath(bundle_dir)
    if os.path.exists(normalized_bundle_dir):
        _LOG.debug(f'Bundle folder `{normalized_bundle_dir}` exists locally.')
        return True
    return False


def upload_bundle_to_s3(bundle_name, force):
    if if_bundle_exist(bundle_name) and not force:
        raise ProjectStateError(
            f"Bundle name '{bundle_name}' already exists in deploy bucket. "
            f"Please use another bundle name or delete the bundle"
        )

    bundle_path = resolve_bundle_directory(bundle_name=bundle_name)
    build_meta_path = build_path(bundle_path, BUILD_META_FILE_NAME)
    meta_resources = json.load(open(build_meta_path))
    validate_deployment_packages(bundle_path=resolve_all_bundles_directory(),
                                 meta_resources=meta_resources)
    _LOG.info('Bundle was validated successfully')
    paths = []
    for root, dirs, file_names in os.walk(bundle_path):
        for file_name in file_names:
            paths.append(file_name)
    executor = ThreadPoolExecutor(max_workers=10)
    futures = []
    for path in paths:
        if 'output/' not in path:
            path_to_package = build_path(bundle_path, path)
            _LOG.debug('Going to upload file: {0}'.format(path_to_package))
            arg = {
                'path': build_path(bundle_name, path),
                'path_to_package': path_to_package
            }
            futures.append(executor.submit(_put_package_to_s3, arg))
    return futures


def create_bundles_bucket():
    from syndicate.core import CONFIG, CONN
    if CONN.s3().is_bucket_exists(CONFIG.deploy_target_bucket):
        USER_LOG.warning(
            f"Bundles bucket '{CONFIG.deploy_target_bucket}' already exists"
        )
        return False
    else:
        _LOG.info(
            'Bundles bucket {0} does not exist. Creating bucket..'.format(
                CONFIG.deploy_target_bucket))
        CONN.s3().create_bucket(
            bucket_name=CONFIG.deploy_target_bucket,
            location=CONFIG.region)
        CONN.s3().put_public_access_block(CONFIG.deploy_target_bucket)
        _LOG.info('{0} bucket created successfully'.format(
            CONFIG.deploy_target_bucket))
    return True


def load_bundle(bundle_name, src_account_id, src_bucket_region,
                src_bucket_name, role_name):
    from syndicate.core import CONFIG, RESOURCES_PROVIDER
    assert_bundle_bucket_exists()
    try:
        _LOG.debug(
            'Going to assume {0} role from {1} account'.format(role_name,
                                                               src_account_id))
        credentials = RESOURCES_PROVIDER.sts().get_temp_credentials(
            role_name, src_account_id, 3600)
        access_key = credentials['AccessKeyId']
        secret_key = credentials['SecretAccessKey']
        session_token = credentials['SessionToken']
        src_s3_conn = S3Connection(region=src_bucket_region,
                                   aws_access_key_id=access_key,
                                   aws_secret_access_key=secret_key,
                                   aws_session_token=session_token)
        _LOG.debug('Credentials were assumed successfully')
    except ClientError:
        raise ConfigurationError(
            f"Cannot assume '{role_name}' role. Please verify that the role "
            f"exists and has correct trusted relationships to be assumed from "
            f"'{CONFIG.account_id}' account."
        )
    if not src_s3_conn.is_bucket_exists(src_bucket_name):
        raise ConfigurationError(
            f"'{src_account_id}' account does not have a '{src_bucket_name}' "
            f"bucket. Please verify that you have configured the correct "
            f"bucket name.")
    _LOG.info('Going to find S3 keys for bundle: {0}'.format(bundle_name))
    objects = src_s3_conn.list_objects(bucket_name=src_bucket_name,
                                       prefix=bundle_name)
    artifacts_names = [meta['Key'] for meta in objects]
    _LOG.info('Found {0} artifacts: {1}'.format(len(artifacts_names),
                                                artifacts_names))

    bundle_path = build_path(CONFIG.project_path, ARTIFACTS_FOLDER,
                             bundle_name)
    for dirpath, dirnames, files in os.walk(bundle_path):
        if files:
            raise ProjectStateError(
                'Bundle name is already exists. Please, verify that have '
                'configured the correct bundle name.'
            )

    # TODO create_pool can be used
    executor = ThreadPoolExecutor(max_workers=10)
    futures = []
    for key in artifacts_names:
        arg = {
            'conn': src_s3_conn,
            'bucket_name': src_bucket_name,
            'key': key,
            'path': build_path(CONFIG.project_path, ARTIFACTS_FOLDER, key)
        }
        futures.append(executor.submit(_download_package_from_s3, arg))
    return futures


@unpack_kwargs
def _download_package_from_s3(conn, bucket_name, key, path):
    conn.download_file(bucket_name, key, path)


@unpack_kwargs
def _put_package_to_s3(path, path_to_package):
    from syndicate.core import CONN, CONFIG
    key_compound = PurePath(CONFIG.deploy_target_bucket_key_compound,
                            path).as_posix()
    CONN.s3().upload_single_file(path_to_package, key_compound,
                                 CONFIG.deploy_target_bucket)


def remove_bundle_dir_locally(bundle_name: str, force_upload: bool):
    if if_bundle_exist_locally(bundle_name) and not force_upload:
        raise ProjectStateError(
            f'Bundle name \'{bundle_name}\' already exists locally. '
            f'Please use another bundle name or delete the existing'
        )
    if force_upload:
        _LOG.info(f'Force upload is enabled, going to check if bundle '
                  f'directory already exists locally.')
        bundle_dir = resolve_bundle_directory(bundle_name=bundle_name)
        normalized_bundle_dir = os.path.normpath(bundle_dir)
        if os.path.exists(normalized_bundle_dir):
            _LOG.warning(f'Going to remove bundle folder '
                         f'`{normalized_bundle_dir}` locally.')
            try:
                shutil.rmtree(normalized_bundle_dir)
            except Exception as e:
                _LOG.error(f'Cannot delete folder {normalized_bundle_dir}')
                raise e
